# -*- coding: utf-8 -*-
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
# This is a python translation of the ACSIS chartplt.cc code
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


import sys

from qwt.qt.QtGui import QPen, QBrush, QColor,QWidget, QFont, QFontInfo,QMessageBox
from qwt.qt.QtGui import QApplication, QHBoxLayout, QFrame, QLabel
from qwt.qt.QtCore import Qt, QObject, pyqtSignal, QTimer, QPoint
from qwt import QwtPlot, QwtPlotMarker, QwtPlotCurve, QwtText

from .QwtSpy_qt5 import *
from . import zoomwin_qt5

import random
import numpy
try:
  from unidecode import unidecode
except:
  pass

HAS_TIMBA = False
try:
  from MeqGUI.GUI.pixmaps import pixmaps
  HAS_TIMBA = True
except:
  pass

class ChartPlot(QWidget):
  auto_offset_value = pyqtSignal(float)

  def __init__(self,control_menu=None,num_curves=None,parent=None,name=None):
    """ Collects information about timing from setup_parameters dict.
        Resizes and initialises all the variables.  
        creates bottom frame and the chart plot
        Does all the connections
    """
    QWidget.__init__(self, parent)

    self.setMouseTracking(True)
    #Constant strings
    self._zoomInfo = "Zoom: Press mouse button and drag"
    self._cursorInfo = "Cursor Pos: Press middle mouse button in plot region"

    self._plotter_is_active = True
    self._auto_offset = True
    self._ref_chan= 0
    self._offset = 0
    self._source = None
    self._max_range = -10000
    self._data_label = None
    self.info_marker = None
    self._plot_label = None
    self._complex_marker = None
    self._zoom_png_number = 0
    self.has_nans_infs = False
    
    # menu is a ControlMenu object from DataDisplayMainWindow
    self._menu = control_menu;
    
    # constants for complex coponents
    try:
      self.AMP = self._menu.AMP;
      self.PHASE = self._menu.PHASE;
      self.REAL = self._menu.REAL;
      self.IMAG = self._menu.IMAG;
      self._complex_component = self._menu.complex_component;
      #print('selection parameters ', self.AMP, self.PHASE, self.REAL, self.IMAG)
    except:
      print('This script can only be run indirectly via the DataDisplayMainWindow script')
      sys.exit()
# create plotter
    self._plotter = QwtPlot(self)

# set plotter fonts to same size as in display_image.py code
    self.title_font = QFont(QApplication.font());
    fi = QFontInfo(self.title_font);
    # and scale it down to 70%
    self.title_font.setPointSize(fi.pointSize()*0.7);

    #Create the plot widget
    self._y_title = QwtText("Value (Relative Scale)")
    self._y_title.setFont(self.title_font)
    self._x_title = QwtText("Time Sequence (Relative Scale)")
    self._x_title.setFont(self.title_font)
    self._plotter.setAxisTitle(QwtPlot.yLeft, self._y_title)
    self._plotter.setAxisTitle(QwtPlot.xBottom, self._x_title)


    #Set the background color
    self._plotter.setCanvasBackground(Qt.white)
    
    # we seem to need a layout for PyQt
    self.box1 = QHBoxLayout(self)
    self.box1.addWidget(self._plotter)

    # set a printer facility
#   self.plotPrinter = plot_printer_qt4.plot_printer(self._plotter)

    #Number of curves/beams
    if not num_curves is None:
      self._nbcrv = num_curves
      # make sure _nbcrv is a multiple of four
      nbcrv_bin = 4
      divis = self._nbcrv // nbcrv_bin
      if divis * nbcrv_bin < self._nbcrv:
        divis = divis + 1
        self._nbcrv = divis * nbcrv_bin
    else:
      self._nbcrv = 16

    #Initialization of the size of the arrays to put the curve in
    self._ArraySize = 6
    self._x_displacement = self._ArraySize // 6 
    self.set_x_axis_sizes()

    # this is the label of the vells that is being displayed
    self._data_index = None;
    
    # Initialize all the arrays containing the curve data 
    # code for this?

    self._closezoom = False
    self._mainpause = False
    
    #initialize the mainplot zoom variables
    self._d_zoomActive = self._d_zoom = False
    self._is_vector = False
    self.source_marker = None
	    
    #initialize zoomcount.  Count the number of zoom windows opened
    self._zoomcount = 0

    # set up pop up text
    font = QFont("Helvetica",10)
    self.white = QColor(255,255,255)
    self._popup_text = QLabel(self)
    self._popup_text.setFont(font)
    self._popup_text.setStyleSheet("QWidget { background-color: %s }" % self.white.name())
    self._popup_text.setFrameStyle(QFrame.Box | QFrame.Plain)
    # start the text off hidden at 0,0
    self._popup_text.hide()


    # timer to allow redisplay
    self._display_interval_ms = 2000
    self._display_refresh = QTimer(self)
    self._display_refresh.start(self._display_interval_ms)
    self._refresh_flag = True
    try:
      self._complex_component = self.AMP;
    except:
      print('This script can only be run indirectly via the DataDisplayMainWindow script')
      sys.exit()
    

#   self._mainwin = parent and parent.topLevelWidget()
    self._mainwin = parent

    # connect menu actions
    
    self._menu.close_window.triggered.connect(self.quit)
    
    self._menu.reset_zoomer.triggered.connect(self.reset_zoom)

# the following is commented out until postscript/pdf printing works 
# properly with Qt 4 widgets
#   self._print = menu.addAction("Print",self.plotPrinter.do_print)

    self._menu.clear_plot.triggered.connect(self.clear_plot)

    self._menu.close_popups.triggered.connect(self.closezoomfun)

    self._menu.show_flagged_data.triggered[bool].connect(self.change_flag_parms)

    #***** add Fixed scale vs Offset scale submenu
    self._menu.autoscale.triggered[bool].connect(self.change_scale_type)

    self._menu.offset_value.triggered.connect(self.change_offset_value)
    
    self._menu.show_labels.triggered[bool].connect(self.enable_labels)

    self._menu.changeComplexComponent.connect(self.update_complex_selector)
    self._menu.changeVellsComponent.connect(self.update_vells_selector)
    self._menu.changeStokesComponent.connect(self.update_stokes_selector)
    self.spy = Spy(self._plotter.canvas())
    self.prev_xpos = None
    self.prev_ypos = None
    self.xzoom_loc = None
    self.yzoom_loc = None
    self.zoomStack = []
    self.zoom_outline = QwtPlotCurve()


    ########### Connections for Signals ############
    self.spy.MouseMove.connect(self.plotMouseMoved)
    self.spy.MousePress.connect(self.plotMousePressed)
    self.spy.MouseRelease.connect(self.plotMouseReleased)


    # redisplay
    self._display_refresh.timeout.connect(self.refresh_event)

    # construct plot / array storage structures etc
    self.createplot(True)
    
  def set_plotter_active():
    self._plotter_is_active = True

  def set_plotter_inactive():
    self._plotter_is_active = False

  def setSource(self, source_string):
    self._source = source_string

  def set_x_axis_sizes(self):
    """ changes sizes and values of arrays used to store x-axis 
        values for plot
    """
    self._x1 = numpy.zeros((self._ArraySize,), numpy.float32)
    self._x2 = numpy.zeros((self._ArraySize,), numpy.float32)
    self._x3 = numpy.zeros((self._ArraySize,), numpy.float32)
    self._x4 = numpy.zeros((self._ArraySize,), numpy.float32)

    # layout parameter for x_axis offsets 
    if self._ArraySize > 6 * self._x_displacement:
      self._x_displacement = self._ArraySize // 6
    for i in range(self._ArraySize):
      self._x1[i] = float(i)
      self._x2[i] = self._ArraySize + self._x_displacement +i
      self._x3[i] = 2 * (self._ArraySize + self._x_displacement) + i
      self._x4[i] = 3 * (self._ArraySize + self._x_displacement) + i

  def closestCurve(self, pos):
        """ from Gerard Vermeulen's EventFilterDemo.py example """
        found, distance, point, index = None, 1e100, -1, -1
        counter = -1
        for curve in self._plotter.itemList():
            try:
              if isinstance(curve, QwtPlotCurve):
                counter = counter + 1
                i, d = curve.closestPoint(pos)
                if i >= 0 and d < distance:
                  index = counter 
                  found = curve
                  point = i
                  distance = d
            except:
              pass

        if found is None:
          return (None, None, None)
        else:
          s = found.sample(point)
          x = s.x()
          y = s.y()
          return (index, x, y, point)
    # closestCurve

  def refresh_plot (self):
    """ causes entire plot to be redone -- called from methods that change the plotted data""";
    self._auto_offset = True
    self._offset = 0
    self._max_range = -10000
    for channel in range(self._nbcrv):
      self._updated_data[channel] = True
    self.reset_zoom()
    self._refresh_flag = True
    self.refresh_event()


  def update_complex_selector (self,label):
    """ callback to handle events from the complex data selection
        sub-context menu 
    """
    # somewhere label has been converted to unicode from ASCII 
    self._complex_component = unidecode(label);
    self._y_title.setText("%s (Relative Scale)"%label);
    self._plotter.setAxisTitle(QwtPlot.yLeft, self._y_title)
    for channel in range(self._nbcrv):
      self._start_offset_test[channel][self._data_index] = 0
    self.refresh_plot();
    return True

  def update_vells_selector (self,label):
    self._data_index = label;
    for channel in range(self._nbcrv):
      self._crv_key[channel].detach()
    self.refresh_plot();

  def update_stokes_selector (self,label):
    print('in update_stokes_selector - no valid response at present so ignored')
#   self._data_index = label;
    for channel in range(self._nbcrv):
      self._crv_key[channel].detach()
    self.refresh_plot();


  def change_scale_type (self,autoscale):
    # click means change to fixed scale
    if autoscale:
      self._plotter.setAxisAutoScale(QwtPlot.yLeft)
    else: 
      # find current data min and max
      scale_max = self._spec_grid_offset.max()
      scale_min = self._spec_grid_offset.min()

      scale_diff = scale_max - scale_min
      scale_max = scale_max + 0.2 * scale_diff
      scale_min = scale_min - 0.2 * scale_diff 

      AxisParms = ScaleSelector(scale_max, scale_min, self)
      AxisParms.scale_values.connect(self.set_scale_values)
      AxisParms.cancel.connect(self.cancel_scale_request)

  def change_offset_value (self):
    OffSetParam = OffsetSelector(self)
    OffSetParam.offset_value.connect(self.set_offset_value)
    OffSetParam.cancel_offset.connect(self.cancel_offset_request)

  def enable_labels (self,enable):
    for channel in range(self._nbcrv):
      self._start_offset_test[channel][self._data_index] = 0
    self.refresh_plot();

  def set_scale_values(self, max_value, min_value):
    if not self._menu.autoscale.isChecked():
      self._plotter.setAxisScale(QwtPlot.yLeft, min_value, max_value)
      if self._plotter_is_active:
        self._plotter.replot()

  def cancel_scale_request(self):
    self._menu.autoscale.setChecked(False);
    self._plotter.setAxisAutoScale(QwtPlot.yLeft)

  def set_offset_value(self, offset_value):
    self._offset = offset_value
    self._refresh_flag = True
    if self._offset < 0.0:
      self._auto_offset = True
      self._offset = 0
      self._max_range = -10000
    else: 
      self._auto_offset = False

  def change_flag_parms (self,*dum):
    self._menu.autoscale.setChecked(True);
    for channel in range(self._nbcrv):
      self._start_offset_test[channel][self._data_index] = 0
    self.refresh_plot();
    
  def reset_zoom (self):
    """ resets data display so all data are visible """
    self._plotter.setAxisAutoScale(QwtPlot.yLeft)
    self._plotter.setAxisAutoScale(QwtPlot.xBottom)
    if self._plotter_is_active:
      self._plotter.replot()
    self._menu.reset_zoomer.setVisible(False)

  def clear_plot(self):
    """ clear the plot of all displayed data """
    # remove any markers
    for i in self._plotter.itemList():
        if isinstance(i, QwtPlotMarker):
          i.detach()
    # remove current curves
    for i in self._plotter.itemList():
        if isinstance(i, QwtPlotCurve):
          i.detach()
    # undo any zooming
    self.reset_zoom()
    self._ArraySize = 6
    self._x_displacement = self._ArraySize // 6 
    self.set_x_axis_sizes()
    self.createplot()

  def dataLabel (self):
    return self._data_label;
  
  def setDataLabel(self, data_label):
    self._data_label = data_label
#    title = self._data_label + " " + self._x_title.text()
#    self._x_title.setText(title)
    self._plotter.setAxisTitle(QwtPlot.xBottom, self._x_title)

  def setPlotLabel(self, plot_label):
    self._plot_label = plot_label

  def destruct_chartplot(self):
    """	turn off global mouse tracking """
    QApplication.setGlobalMouseTracking(False)	
    QWidget.setMouseTracking(False) 
			
  def createplot(self,first_time=False):
    """ Sets all the desired parameters for the chart plot """
    if first_time:
      self._chart_data = {}
      self._flag_data = {}
      self._start_offset_test = {}
      self._updated_data = {}
      self._pause = {}
      self._mrk = {}
      self._position = {}
      self._Zoom = {}
      self._zoom_title = {}
      self._zoom_pen = {}
      self._main_pen = {}
      self._indexzoom = {}
      self._crv_key = {}
      self._source_marker = {}
      for i in range (self._nbcrv):
          self._updated_data[i] = False
          self._indexzoom[i] = False
          self._pause[i] = False
          self._Zoom[i] = None
          self._source_marker[i] = None
          self._mrk[i] = 0
          self._position[i] = ""
          self._zoom_title[i] = "Data for Chart " + str(i)
          self._zoom_pen[i] = Qt.yellow
          self._main_pen[i] = Qt.black
          self._crv_key[i] = QwtPlotCurve("Chart " + str(i))
          self._crv_key[i].setPen(QPen(self._main_pen[i]))
          self._crv_key[i].attach(self._plotter)
          
          self._chart_data[i] = {}
          self._flag_data[i] = {}
          self._start_offset_test[i] = {}
    else:
      for i in range (self._nbcrv):
          self._updated_data[i] = False
          self._pause[i] = False
          self._source_marker[i] = None
          self._mrk[i] = 0
          self._position[i] = ""
          self._main_pen[i] = Qt.black
          self._crv_key[i] = QwtPlotCurve("Chart " + str(i))
          self._crv_key[i].setPen(QPen(self._main_pen[i]))
          self._crv_key[i].attach(self._plotter)
          self._chart_data[i] = {}
          self._flag_data[i] = {}
          self._start_offset_test[i] = {}

  def infoDisplay(self):
    """ Display text under cursor in plot
        Figures out where the cursor is, generates the appropriate text, 
        and puts it on the screen.
    """
    closest_curve, xVal, yVal, data_point = self.closestCurve(self.position_raw)

    #get value of where the mouse was released
    p2x = self._plotter.invTransform(QwtPlot.xBottom, self.xpos_raw)

# determine 'actual' x-axis position
    ref_point=0
    if (p2x >= 0) and (p2x < self._ArraySize):
      ref_point = int(p2x)
    elif p2x >= self._ArraySize + self._x_displacement and p2x < (2*self._ArraySize) + self._x_displacement: 
      ref_point = int(p2x - self._ArraySize - self._x_displacement)
    elif p2x >= 2*(self._ArraySize + self._x_displacement) and p2x < (3*self._ArraySize)+(2*self._x_displacement):
      ref_point = int(p2x - (2*(self._ArraySize + self._x_displacement)))
    elif p2x >= 3*(self._ArraySize+self._x_displacement) and p2x < (4*self._ArraySize) + 3*self._x_displacement: 
      ref_point = int(p2x - (3*(self._ArraySize+self._x_displacement)))

    if self._complex_component == self.PHASE:
      temp_off = ((closest_curve) % (self._nbcrv//4) + 0.5 ) * self._offset
    else:
      temp_off = ((closest_curve) % (self._nbcrv//4)) * self._offset
    ref_value = yVal - temp_off

    message = self.reportCoordinates(ref_point, ref_value, closest_curve)
    message = message +', data point: ' + str(data_point)

    self._popup_text.setText(message)
    self._popup_text.adjustSize()
    try:
      yhb = self._plotter.transform(QwtPlot.yLeft, self._plotter.axisScaleDiv(QwtPlot.yLeft).hBound())
      ylb = self._plotter.transform(QwtPlot.yLeft, self._plotter.axisScaleDiv(QwtPlot.yLeft).lBound())
      xhb = self._plotter.transform(QwtPlot.xBottom, self._plotter.axisScaleDiv(QwtPlot.xBottom).hBound())
      xlb = self._plotter.transform(QwtPlot.xBottom, self._plotter.axisScaleDiv(QwtPlot.xBottom).lBound())
    except:
      yhb = self._plotter.transform(QwtPlot.yLeft, self._plotter.axisScaleDiv(QwtPlot.yLeft).upperBound())
      ylb = self._plotter.transform(QwtPlot.yLeft, self._plotter.axisScaleDiv(QwtPlot.yLeft).lowerBound())
      xhb = self._plotter.transform(QwtPlot.xBottom, self._plotter.axisScaleDiv(QwtPlot.xBottom).upperBound())
      xlb = self._plotter.transform(QwtPlot.xBottom, self._plotter.axisScaleDiv(QwtPlot.xBottom).lowerBound())
    height = self._popup_text.height()
    if self.ypos + height > ylb:
      ymove = self.ypos - 1.5 * height
    else:
      ymove = self.ypos + 0.5 * height
    width = self._popup_text.width()
    if self.xpos + width > xhb:
      xmove = xhb - width
    else:
      xmove = self.xpos
    self._popup_text.move(xmove, ymove)
    if not self._popup_text.isVisible():
      self._popup_text.show()
    

  def closezoomfun(self):

    """ Set closezoom flag active to close all the zoom windows 
        once the user clicked
        the close_all_zoom button from the warning
        put closezoom to 1 and replot
    """
    self._closezoom = True
    self.ReplotZoom()

# close any zoom pop up windows
  def quit(self):
   for curve_no in range(self._nbcrv):
     if self._indexzoom[curve_no]:  
       if not self._Zoom[curve_no] is None:
         self._Zoom[curve_no].Closing()
   self.close()

  def ReplotZoom(self):
    """ If no zoom window opened, sets the flag back to 0
        Looks in the array to see which zooms are opened
        if the closezoom is set to 1
        Close the zoom
        set the flag in the array to 0
        removes 1 from the number of zoom opened
        else
        replot the zoom
    """

    #If no zoom window opened, sets the flag back to 0
    if self._zoomcount==0:
      self._closezoom = False
      
    #Looks in the array to see which zooms are opened
    for curve_no in range(self._nbcrv):
      if self._indexzoom[curve_no] and not self._Zoom[curve_no] is None:
        if self._closezoom:
          #Close the zoom
          self._Zoom[curve_no].close()
          #set the flag in the array to 0
          self._indexzoom[curve_no] = False
          #removes 1 from the number of zoom opened
          self._zoomcount = self._zoomcount - 1
        elif not self._pause[curve_no]:  #replot the zoom
          chart = numpy.array(self._chart_data[curve_no][self._data_index])
          flags = numpy.array(self._flag_data[curve_no][self._data_index])
          self._Zoom[curve_no].update_plot(chart,flags)
          self._Zoom[curve_no]._plotter.replot()

    if self._closezoom:
      self._menu.close_popups.setVisible(False)

  def plotMouseMoved(self, e):
    position = e.pos()
    if self._popup_text.isVisible():
      self._popup_text.hide()
    self.xpos_raw = position.x()
    self.ypos_raw = position.y()
    self.xpos = self._plotter.invTransform(QwtPlot.xBottom, self.xpos_raw)
    self.ypos = self._plotter.invTransform(QwtPlot.yLeft, self.ypos_raw)

    self.position_raw = QPoint(self.xpos_raw,self.ypos_raw)
    if not self.xzoom_loc is None:
      self.xzoom_loc = [self.press_xpos, self.press_xpos,  self.xpos, self.xpos,self.press_xpos]
      self.yzoom_loc = [self.press_ypos, self.ypos,  self.ypos, self.press_ypos,self.press_ypos]
      self.zoom_outline.setData(self.xzoom_loc,self.yzoom_loc)
      if self._plotter_is_active: 
        self._plotter.replot()

  def mapMouseButtons (self,e):
      """Maps a mouse event to one of three buttons.
      To support victims of Jobs, Shift-LeftClick maps to MidClick, and Ctrl+LeftClick maps to RightClick""";
      if e.button() == Qt.LeftButton:
        if e.modifiers()&Qt.ShiftModifier:
          return Qt.MidButton;
        elif e.modifiers()&Qt.ControlModifier:
          return Qt.RightButton;
      return e.button();

  def plotMousePressed(self, e):
      """ callback to handle MousePressed event """
      self._mouse_press = self.mapMouseButtons(e);
      if self._mouse_press == Qt.LeftButton:
        self.infoDisplay();
        self.press_xpos = self.xpos
        self.press_ypos = self.ypos
        self.press_xpos_raw = e.pos().x()
        self.press_ypos_raw = e.pos().y()
        self.xzoom_loc = [self.press_xpos]
        self.yzoom_loc = [self.press_ypos]
        self.zoom_outline.attach(self._plotter)
        if self.zoomStack == []:
          try:
            self.zoomState = (
              self._plotter.axisScaleDiv(QwtPlot.xBottom).lBound(),
              self._plotter.axisScaleDiv(QwtPlot.xBottom).hBound(),
              self._plotter.axisScaleDiv(QwtPlot.yLeft).lBound(),
              self._plotter.axisScaleDiv(QwtPlot.yLeft).hBound(),
              )
          except:
            self.zoomState = (
              self._plotter.axisScaleDiv(QwtPlot.xBottom).lowerBound(),
              self._plotter.axisScaleDiv(QwtPlot.xBottom).upperBound(),
              self._plotter.axisScaleDiv(QwtPlot.yLeft).lowerBound(),
              self._plotter.axisScaleDiv(QwtPlot.yLeft).upperBound(),
              )
      elif self._mouse_press == Qt.RightButton:
        e.accept()
        self._menu.popup(e.globalPos());
    # plotMousePressed()

  def plotMouseReleased(self, e):
      """ callback to handle MouseReleased event """
      if self._mouse_press == Qt.MidButton:
        closest_curve, xVal, yVal, data_point = self.closestCurve(self.position_raw)
        self.zoomcrv(closest_curve)
 #     elif self._mouse_press == Qt.RightButton:
 #       print "accepting release event";
 #       e.accept()
 #       self._menu.popup(e.globalPos());
      elif self._mouse_press == Qt.LeftButton:
        if self._popup_text.isVisible():
          self._popup_text.hide()
# assume a change of <= 2 screen pixels is just due to clicking
# left mouse button for no good reason
        if abs(self.press_xpos_raw - e.pos().x()) > 2 and abs(self.press_ypos_raw - e.pos().y()) > 2:
          xmin = min(self.xpos, self.press_xpos)
          xmax = max(self.xpos, self.press_xpos)
          ymin = min(self.ypos, self.press_ypos)
          ymax = max(self.ypos, self.press_ypos)
          if xmin == xmax or ymin == ymax:
              return
          self.zoomStack.append(self.zoomState)
          self.zoomState = (xmin, xmax, ymin, ymax)
          self._plotter.setAxisScale(QwtPlot.xBottom, xmin, xmax)
          self._plotter.setAxisScale(QwtPlot.yLeft, ymin, ymax)
          self._menu.reset_zoomer.setVisible(True)
          if self._plotter_is_active: 
            self._plotter.replot()
        if not self.xzoom_loc is None:
          self.zoom_outline.detach()
          self.xzoom_loc = None
          self.yzoom_loc = None

    # plotMouseReleased()

  def reportCoordinates(self, x, y, crv):
    """Format mouse coordinates as real world plot coordinates.
    """
    if not self._plot_label is None:
      plot_label = self._plot_label[crv+self._ref_chan] + ': '
    else:
      plot_label = str(crv+self._ref_chan) + ': '
    temp_str = "nearest x=%-.3g" % x
    temp_str1 = " y=%-.3g" % y
    message = plot_label + temp_str + temp_str1 
    return message
    # reportCoordinates()

  def zoomcrv(self, crv):
    """ If one tries to open a zoom of a curve that is already opened
        open a warning
        Increase the count of the number of curve opened
        Put a flag to indicate wich zoom is opened
        Open a zoompopup of the chosen curve
    """
    if self._indexzoom[crv]:
      if not self._Zoom[crv] is None:
        self._Zoom[crv].show()
        Message = "A zoom of this curve is already opened"
        zoom_message = QMessageBox.warning(self, "ChartPlot", Message)
#       self._Zoom[crv].raise()
    else:
      #To know how many zoom windows opened (so +1)
      self._zoomcount = self._zoomcount + 1
  
      #Get the color of the curve
      pen = self._crv_key[crv].pen()
  
      #Put a flag to indicate wich zoom is open (for reploting)
      self._indexzoom[crv] = True

      #open a zoom of the selected curve
      PlotArraySize = self._x1.size
      chart = numpy.array(self._chart_data[crv][self._data_index])
      flags = numpy.array(self._flag_data[crv][self._data_index])
      
      self._Zoom[crv] = zoomwin_qt5.ZoomPopup(crv+self._ref_chan, self._x1, chart, flags, pen)
      if not self._source is None:
        self._Zoom[crv].setWindowTitle(self._source)
        
      if not self._data_label is None:
        if not self._plot_label is None:
          plot_label = self._plot_label[crv+self._ref_chan]
        else:
          plot_label = None
        self._Zoom[crv].setDataLabel(self._data_label, plot_label, self._is_vector)
      self._pause[crv] = False
      self._Zoom[crv].winclosed.connect(self.myindex)
      self._Zoom[crv].winpaused.connect(self.zoomPaused)
      self._Zoom[crv].save_zoom_display.connect(self.grab_zoom_display)

      # make sure option to close all Popup windows is seen
      self._menu.close_popups.setVisible(True)


  def grab_zoom_display(self, title, crvtemp):
    self._zoom_png_number = self._zoom_png_number + 1
    png_str = '_' + str(self._zoom_png_number)
    if not self._source is None:
      if title is None:
        save_file = self._source + '_' + './meqbrowser.png'
      else:
        save_file = self._source + '_' + title + png_str + '.png'
    else:
      if title is None:
        save_file = './meqbrowser.png'
      else:
        save_file = title + png_str + '.png'
    save_file_no_space= save_file.replace(' ','_')
    result = QPixmap.grabWidget(self._Zoom[crvtemp-self._ref_chan]).save(save_file_no_space, "PNG")

  def set_offset(self,parameters=None):
    """ Update the display offset.
    """
    if not parameters is None:
      if parameters.haskey("default_offset"):
        self._offset = parameters.get("default_offset")
        self._refresh_flag = True
        if self._offset < 0.0:
          self._auto_offset = True
          self._offset = 0
          self._max_range = -10000
        else:
          self._auto_offset = False

  def set_offset_scale(self,new_scale):
    """ Update the display offset.
    """
    self._offset = new_scale
    self._refresh_flag = True
    if self._offset < 0.0:
      self._auto_offset = True
      self._offset = 0
      self._max_range = -10000
    else:
      self._auto_offset = False

    for channel in range(self._nbcrv):
      self._updated_data[channel] = True
      self._start_offset_test[channel][self._data_index] = 0
    self.reset_zoom()

  def set_auto_scaling(self):
    """ Update the display offset.
    """
    self._auto_offset = True
    self._offset = 0
    self._refresh_flag = True
    self._max_range = -10000
    for channel in range(self._nbcrv):
      self._updated_data[channel] = True
      self._start_offset_test[channel][self._data_index] = 0
    self.reset_zoom()

  def ChartControlData(self):
    """ Resize array if no zoom windows are open.
    """
    #Works if no zoom windows are open
    if self._zoomcount == 0:
      resizex(self._ArraySize)
    else:
      Message = "All zoom windows should be closed\nto perform action."
      zoom_message = QMessageBox.warning(self, "ChartPlot", Message)

# this is the equivalent of zoomwarn2.py - use this instead if and when ...
#   Message = "Please put the offset at its maximum value\nbefore zooming by a click on the plot"
#   mb_reporter = QMessageBox.information(self, "ChartPlot", Message)

  def resizex(self, size):
    """ Get the size the arrays should have
        reinitialize the arrays
    """
      
    for i in range(self._nbcrv):
      if self._indexzoom[i] and not self._Zoom[i] is None:
        self._Zoom[i].resize_x_axis(self._ArraySize)
    # y data with offset for plotting on the main display
    self._spec_grid_offset.resize(self._ArraySize, self._nbcrv+1)
    self._spec_grid_offset = 0
    self._spec_ptr.resize(self._nbcrv+1)
    for i in range(self._nbcrv):
      self._spec_ptr[i] = dtemp + self._ArraySize * i 

    # the y values for zoom pop up windows without the offset
    self._spec_grid.resize(self._ArraySize,self._nbcrv+1)
    self._spec_grid = 0
    self._spec_ptr_actual.resize(self._nbcrv+1)
    for i in range(self._nbcrv):
      self._spec_ptr_actual[i] = dtemp + self._ArraySize * i 

  def zoomCountmin(self):
    """ Reduces the count of zoom windows opened
    """
    self._zoomcount = self._zoomcount - 1

  def zoomPaused(self, crvtemp):
    self._pause[crvtemp-self._ref_chan] = not self._pause[crvtemp-self._ref_chan]


  def myindex(self, crvtemp):
    """ Sets the indexzoom flag of the curve that just closed to False
    """
    self._indexzoom[crvtemp-self._ref_chan] = False
    self._Zoom[crvtemp-self._ref_chan].exec_close()

  def updateEvent(self, data_dict):
    """ Update the curve data for the given channel
    """
    # in updateEvent, channel goes from 0 to _nbcrv - 1
    # ---
    # if necessary, first remove old data from front of queue
    # add new data to end of queue

    # do we have an incoming dictionary?
    q_pos_str = "Sequence Number " + str( data_dict['sequence_number'])
    new_chart_flags = None
    try:
      self._source = data_dict['source']
      new_chart_flags = data_dict['flags']
    except:
      self._source = None
      pass
    channel_no = data_dict['channel']
    new_chart_val = data_dict['value']
    has_keys = True
    add_vells_menu = False
    try:
      data_keys = list(new_chart_val.keys())
      add_vells_menu = True
    except:
      has_keys = False
      data_keys = []
      data_keys.append(0)
#    print "Data keys",data_keys;
      
    # set current index to first vells, if not initialized
    self._data_index = self._menu.vells_component;
    if self._data_index is None:
      self._data_index = data_keys[0];

    if channel_no >=  self._nbcrv:
      factor = int (channel_no / self._nbcrv)
      self._ref_chan = factor * self._nbcrv
      channel = channel_no - self._ref_chan
    else:
      channel = channel_no
      
    for keys in data_keys: 
      if keys not in self._chart_data[channel]:
        self._chart_data[channel][keys] = []
        self._flag_data[channel][keys] = []
        self._start_offset_test[channel][keys] = 0
#        if len(data_keys) > 1:
#          self._menu.add_vells_key(keys);

      incoming_data = None
      incoming_flags = None
      if has_keys:
        try:
          incoming_data = new_chart_val[keys]
          incoming_flags = new_chart_flags[keys]
        except:
          pass
      else:
        incoming_data = new_chart_val
        incoming_flags = new_chart_flags

      # skip rest of processing if incoming_data is non-existent
      if incoming_data is None:
        continue
      else:
        self._refresh_flag = True

      #print 'incoming flags ', incoming_flags
# first, do we have a scalar?
      is_scalar = False
      scalar_data = 0.0
      flag_data = 0
      try:
        shape = incoming_data.shape
# note: if shape is 1 x N : 1 time point with N freq points
# note: if shape is N x 1 or (N,) : N time points with 1 freq point
#     _dprint(3,'data_array shape is ', shape)
      except:
        is_scalar = True
        scalar_data = incoming_data
        if not incoming_flags is None:
          flag_data = incoming_flags
      if not is_scalar and len(shape) == 1:
        if shape[0] == 1:
          is_scalar = True
          scalar_data = incoming_data[0]
          if not incoming_flags is None:
            flag_data = incoming_flags[0]
          
      if is_scalar:
        #print 'scalar is ', scalar_data
#     if len(self._chart_data[channel]) > self._ArraySize-1:
#       differ = len(self._chart_data[channel]) - (self._ArraySize-1)
#       for i in range(differ):
#         del self._chart_data[channel][0]

        if scalar_data == numpy.inf or scalar_data == numpy.nan:
          flag_data = 10  
        self._chart_data[channel][keys].append(scalar_data)
        self._flag_data[channel][keys].append(flag_data)
        self._updated_data[channel] = True

        # take care of position string
        self._position[channel] = q_pos_str
  
      # otherwise we have an array (assumed to be 1D for the moment), 
      # so we add to the stored chart data in its entirety
      else:
        self._is_vector = True
        num_elements = 1
        for i in range(len(incoming_data.shape)):
          num_elements = num_elements * incoming_data.shape[i]

        flattened_array = incoming_data.copy()
        flattened_array.shape = (num_elements,)
        if num_elements > 1 and incoming_data.shape[0] == 1:
          if  self._data_label is None:
            self._x_title.setText("Frequency Spectrum per Tile Block (Relative Scale)")
          else:
            self._x_title.setText(self._data_label + " Frequency Spectrum per Tile Block (Relative Scale)")
          self._plotter.setAxisTitle(QwtPlot.xBottom, self._x_title)
        # check for NaNs and Infs
        self.has_nans_infs = False
        nan_test = numpy.isnan(flattened_array)
        inf_test = numpy.isinf(flattened_array)
        if nan_test.max() > 0 or inf_test.max() > 0:
          delete = nan_test | inf_test
          keep = ~nan_test & ~inf_test
          self.has_nans_infs = True

        append = self._menu.append.isChecked();
        if append: 
          for i in range(len(flattened_array)):
            self._chart_data[channel][keys].append(flattened_array[i])
        else:
          self._chart_data[channel][keys] = flattened_array
        self._updated_data[channel] = True

        if incoming_flags is None:
          flattened_array = numpy.zeros((num_elements,), numpy.int32)
        else:
          flattened_array = numpy.reshape(incoming_flags.copy(),(num_elements,))
        if self.has_nans_infs:
          flattened_array[delete] = 10
        if append: 
          for i in range(len(flattened_array)):
            self._flag_data[channel][keys].append(flattened_array[i])
        else:
          self._flag_data[channel][keys] = flattened_array
          self._start_offset_test[channel][keys] = 0

      if self._ArraySize < len(self._chart_data[channel][keys]):
        self._ArraySize = len(self._chart_data[channel][keys])
        self.set_x_axis_sizes()

  def refresh_event(self):
    """ redisplay plot and zoom
        if refresh flag is true (new data exists)
        call replots
        set refresh flag False
    """

    # if no changes have been made, just return
    if not self._refresh_flag and not self._plotter_is_active:
      return

    # first determine offsets
    if self._auto_offset:
      for channel in range(self._nbcrv):
        if self._updated_data[channel] and self._data_index in self._chart_data[channel]:
          try:
            chart = numpy.array(self._chart_data[channel][self._data_index])
            #print 'shape chart ', chart.shape, ' ', chart
          except:
            self._updated_data[channel] = False
            pass
          try:
            flags = numpy.array(self._flag_data[channel][self._data_index])
          except:
            self._updated_data[channel] = False
            pass
          # no actual data?
          if chart.shape[0] < 1:
            self._updated_data[channel] = False
          if self._updated_data[channel]:
            test_chart = numpy.compress(flags==0,chart)
            if test_chart.shape[0] > 0:
              if chart.dtype == numpy.complex64 or chart.dtype == numpy.complex128:
                self._menu.showComplexControls(True)
                complex_chart = test_chart.copy()
                self._y_title.setText("%s (Relative Scale)"%self._complex_component)
                self._plotter.setAxisTitle(QwtPlot.yLeft, self._y_title)
                if self._complex_component == self.AMP:
                  cplx_chart = abs(complex_chart)
                elif self._complex_component == self.REAL:
                  cplx_chart = complex_chart.real
                elif self._complex_component == self.IMAG:
                  #print('selecting imaginary')
                  cplx_chart = complex_chart.imag
                elif self._complex_component == self.PHASE:
                  cplx_chart = numpy.angle(complex_chart)
                tmp_max = cplx_chart.max()
                tmp_min = cplx_chart.min()
              else:
                self._menu.showComplexControls(False)
                tmp_max = test_chart.max()
                tmp_min = test_chart.min()
              chart_range = abs(tmp_max - tmp_min)
              # Check if we break any highest or lowest limits
              # this is important for offset reasons.
              # If a chart range is zero, we still need to create
              # an offset so ...
              if chart_range < 1.0e-6:
                chart_range = 1.0e-6
              if chart_range > self._max_range:
                self._max_range = chart_range
              self._start_offset_test[channel][self._data_index] = chart.shape[0]

      #set the max value of the offset
      if 1.1 * self._max_range > self._offset:
        self._offset = 1.1 * self._max_range

        if self._offset < 0.001:
          self._offset = 0.001
        self.auto_offset_value.emit(self._offset)

    # -----------
    # now update data
    # -----------

    for channel in range(self._nbcrv):
      if self._updated_data[channel] and self._data_index in self._chart_data[channel]:
        index = channel+1
        #Set the values and size with the curves
        if index >= 1 and index <= self._nbcrv//4:
          temp_x = self._x1
        elif index >= (self._nbcrv//4)+1 and index <= self._nbcrv//2:
          temp_x = self._x2
        elif index >= (self._nbcrv//2)+1 and index <= 3*(self._nbcrv//4):
          temp_x = self._x3
        elif index >= (3*(self._nbcrv/4))+1 and index <= self._nbcrv:
          temp_x = self._x4

        if self._complex_component == self.PHASE:
          temp_off = (channel % (self._nbcrv//4) + 0.5 ) * self._offset
        else:
          temp_off = (channel % (self._nbcrv//4)) * self._offset

        chart = numpy.array(self._chart_data[channel][self._data_index])
        flags = numpy.array(self._flag_data[channel][self._data_index])
        if chart.dtype == numpy.complex64 or chart.dtype == numpy.complex128:
          complex_chart = chart.copy()
          if self._complex_component == self.AMP:
            self._crv_key[channel].setPen(QPen(Qt.red))
            cplx_chart = abs(complex_chart)
          elif self._complex_component == self.REAL:
            self._crv_key[channel].setPen(QPen(Qt.blue))
            cplx_chart = complex_chart.real
          elif self._complex_component == self.IMAG:
            self._crv_key[channel].setPen(QPen(Qt.gray))
            cplx_chart = complex_chart.imag
          elif self._complex_component == self.PHASE:
            self._crv_key[channel].setPen(QPen(Qt.green))
            cplx_chart = numpy.angle(complex_chart)
          # don't display flagged data
          if not self._menu.show_flagged_data.isChecked():
            x_plot_values = numpy.compress(flags==0,temp_x)
            y_plot_values = numpy.compress(flags==0,cplx_chart)
            if y_plot_values.shape[0] == 0:
              y_plot_values = None
          else:
            # can't display NaNs
            if flags.max() == 10:
              x_plot_values = numpy.compress(flags<10,temp_x)
              y_plot_values = numpy.compress(flags<10,cplx_chart)
            else:
              x_plot_values = temp_x
              y_plot_values = cplx_chart

          if not y_plot_values is None: 
            self._crv_key[channel].setData(x_plot_values , y_plot_values+temp_off)
            self._crv_key[channel].attach(self._plotter)
            ylb = y_plot_values[0] + temp_off 
        else:
          self._crv_key[channel].setPen(QPen(Qt.black))
          # don't display flagged data
          if not self._menu.show_flagged_data.isChecked():
            x_plot_values = numpy.compress(flags==0,temp_x)
            y_plot_values = numpy.compress(flags==0,chart)
            if y_plot_values.shape[0] == 0:
              y_plot_values = None
          else:
            # can't display NaNs
            if flags.max() == 10:
              x_plot_values = numpy.compress(flags<10,temp_x)
              y_plot_values = numpy.compress(flags<10,chart)
            else:
              x_plot_values = temp_x
              y_plot_values = chart

          if not y_plot_values is None: 
            self._crv_key[channel].setData( x_plot_values , y_plot_values+temp_off)
            self._crv_key[channel].attach(self._plotter)
            ylb = y_plot_values[0] + temp_off 

        # update marker with info about the plot
        if not self._source_marker[channel] is None:
          self._source_marker[channel].detach()
        if not y_plot_values is None and self._menu.show_labels.isChecked():
          if not self._plot_label is None:
            message = self._plot_label[channel + self._ref_chan]
          else:
            message =  str(channel + self._ref_chan)
# text marker giving source of point that was clicked
          if self._source_marker[channel] is None:
            self._source_marker[channel] = QwtPlotMarker()
          xlb = temp_x[0]
          self._source_marker[channel].setValue(xlb, ylb)
          self._source_marker[channel].setLabelAlignment(Qt.AlignRight | Qt.AlignTop)
          fn = self._plotter.fontInfo().family()
          text = QwtText(message)
          text.setColor(Qt.blue)
          text.setBackgroundBrush(QBrush(Qt.yellow))
          text.setFont(QFont(fn,7,QFont.Bold,False))
          self._source_marker[channel].setLabel(text)
          self._source_marker[channel].attach(self._plotter)
	 
# need to update any corresponding Zoom window     
#        self._Zoom[channel].resize_x_axis(len(self._chart_data[channel]))
        # Wait that all the channels are modified before 
        # plotting with the new x array


        if not self._mainpause:
          self._refresh_flag = True
	
        #Put the integration number back to 1 for next time
        self._updated_data[channel] = False
      # --------------
      # refresh screen
      # --------------
    if self._refresh_flag: 
      if self._plotter_is_active:
        self._plotter.replot()
      self.ReplotZoom()
      self._refresh_flag = False


  def cancel_offset_request(self):
    return        # do nothing

  def start_test_timer(self, time):
    self.counter = 0
    # stuff for tests
    self.startTimer(time)

  def timerEvent(self, e):
    self.counter = self.counter + 1
    test_array = numpy.zeros((128,), numpy.float32)
    data_dict = {}
    data_dict['source'] = 'abc'
    data_dict['sequence_number'] = self.counter
    data_dict['value'] = test_array

    for i in range(test_array.shape[0]):
      test_array[i] = 0.05 * random.random()
    data_dict['channel'] = 1
    self.updateEvent(data_dict)

    for i in range(test_array.shape[0]):
      test_array[i] = 3 * random.random()
    data_dict['channel'] = 3
    self.updateEvent(data_dict)

    for i in range(test_array.shape[0]):
      test_array[i] = 11 * random.random()
    data_dict['channel'] = 11
    self.updateEvent(data_dict)

    return

def make():
    demo = ChartPlot(None)
    demo.show()
    demo.start_test_timer(500)
    return demo

def main(args):
    app = QApplication(args)
    demo = make()
    app.exec_()


# Admire
if __name__ == '__main__':
    main(sys.argv)





#!/usr/bin/env python3

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#

# these functions create / delete various widgets related
# to plotting. I've tried to collect all such operations
# into one file so that maintenance is easy.

from qwt.qt.QtGui import QApplication,QWidget
from qwt.qt.QtCore import Qt

import numpy

# modules that are imported

HAS_TIMBA = False
try:
  from Timba import utils
  from MeqGUI.GUI import widgets
  from MeqGUI.Plugins.display_image_qt5 import QwtImageDisplay
  from MeqGUI.Plugins.QwtColorBar_qt5 import QwtColorBar
  from MeqGUI.Plugins.ND_Controller_qt5 import ND_Controller

  from Timba.utils import verbosity
  _dbg = verbosity(0,name='plotting_functions');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
  HAS_TIMBA = True
except:
  pass

def create_ColorBar (layout,layout_parent, plotter, plotPrinter):
  """ this function adds a colorbar for 2-D displays """

  # create two color bars in case we are displaying complex arrays
  colorbar = {}
  for i in range(2):
    colorbar[i] =  QwtColorBar(colorbar_number= i, parent=layout_parent)
    colorbar[i].setMaxRange((-1, 1))
    plotter.max_image_range.connect(colorbar[i].handleRangeParms)
    plotter.display_type.connect(colorbar[i].setDisplayType)
    plotter.show_colorbar_display.connect(colorbar[i].showDisplay)
    colorbar[i].set_image_range.connect(plotter.setImageRange)
    if i == 0:
      layout.addWidget(colorbar[i], 0, i)
      colorbar[i].show()
    else:
      layout.addWidget(colorbar[i], 0, 2)
      colorbar[i].hide()
# plotPrinter.add_colorbar(colorbar)
  return colorbar

def delete_2D_Plotters(colorbar, visu_plotter):
  """ delete 2D plotter and associated colorbars """
  if HAS_TIMBA:
    _dprint(3, 'got 3d plot request, deleting 2-d stuff')
  colorbar[0].setParent(QWidget())
  colorbar[1].setParent(QWidget())
  colorbar = {}
  visu_plotter.setParent(QWidget())
  visu_plotter = None
  return visu_plotter

def create_array_selector(plotter, rank, shape, first_axis,second_axis,third_axis=-1):
  """ create subarray selection to be extracted from ND array for display """
  if not plotter is None:
    plotter.delete_cross_sections()
  if first_axis is None:
    first_axis = -1
  if third_axis is None:
    third_axis = -1
  array_selector = []
  for i in range(rank):
    if i == first_axis:
      axis_slice = slice(0,shape[first_axis])
      array_selector.append(axis_slice)
    elif i == second_axis:
      axis_slice = slice(0,shape[second_axis])
      array_selector.append(axis_slice)
    elif i == third_axis:
      axis_slice = slice(0,shape[third_axis])
      array_selector.append(axis_slice)
    else:
      array_selector.append(0)
  return array_selector

def create_2D_Plotters(layout, layout_parent):
  """ create 2D plotter """
  twoD_plotter = QwtImageDisplay(layout_parent)
  layout.addWidget(twoD_plotter, 0, 1)
  twoD_plotter.show()
  return twoD_plotter

def create_ND_Plotter (layout, layout_parent):
  """ create a new ND plotter """
  ND_plotter = vtk_qt_3d_display(layout_parent)
  layout.addWidget(ND_plotter,0,0,1,2)
  ND_plotter.show()
  if HAS_TIMBA:
    _dprint(3, 'issued show call to ND_plotter')
  return ND_plotter

def create_ND_Controls (layout, layout_parent, array_shape, ND_Controls, ND_plotter, labels=None, parms=None, num_axes=2):
  """ this function adds the extra GUI control buttons etc if we are
        displaying data for a Timba.array of dimension 3 or greater 
  """
  if not ND_Controls is None:
    ND_Controls.setParent(QWidget())
    ND_Controls = None
  ND_Controls = ND_Controller(array_shape, labels, parms, num_axes,layout_parent)
  layout.addWidget(ND_Controls,2,0,1,2)
  if ND_Controls.get_num_selectors() > num_axes:
    ND_Controls.showDisplay(1)
  else:
    ND_Controls.showDisplay(0)
    if not ND_plotter is None:
      ND_plotter.HideNDButton()
  if HAS_TIMBA:
    _dprint(3, 'ND_Controls object should appear ', ND_Controls)
  return ND_Controls





# -*- coding: utf-8 -*-
#/usr/bin/env python

#% $Id: display_image.py 6838 2009-03-08 06:21:51Z twillis $


#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import numpy

from qwt.qt.QtGui import QApplication, QPen
from qwt.qt.QtCore import Qt
from qwt import (QwtPlot, QwtPlotMarker, QwtPlotGrid, QwtPlotCurve,
                 QwtScaleMap, QwtScaleDraw, QwtScaleDiv)


class ComplexScaleDraw(QwtScaleDraw):
    """ override default behaviour of QwtScaleDraw class """
    def __init__(self,start_value=None,end_value=None, divisor=None):
        QwtScaleDraw.__init__(self)
        self.do_separator = True
        self.start_value = start_value
        self.end_value = end_value
        self.divisor = divisor
        if not self.start_value is None and not self.end_value is None:
          self.delta = self.end_value - self.start_value
          self.separator = self.end_value
          self.divisor = None
        else:
          self.separator = self.divisor
          self.delta = None
          self.start_value = None
          self.end_value = None

    # __init__()

    def label(self, v):
        """ Override default QwtScaleDraw generation of labels along axis """
        if not self.end_value is None:
          if v >= self.end_value:
              v = v - self.delta
        else:
          if not self.divisor is None:
            v = v % self.divisor
        return QwtScaleDraw.label(self, v)

    def draw_separator(self, separator_flag=True):
        """ test to draw major tick mark at separation location """
        self.do_separator = separator_flag

    def draw(self, painter, dummy):
        # for some reason, in qwt 5 an extra parameter is needed
        # in the above function definition. The corresponding
        # C++ code offers several possibilities, but just a 'dummy'
        # seems adequate ....
        # 1. paint real value ticks as done in C++ code
        # 2. paint imag value ticks with the shift depending on
        #    the dimension and tick spacing
        # 3. draw backbone, if needed

        step_eps = 1.0e-6    # constant value in the c++ equivalent
        majLen = QwtScaleDraw.tickLength(self,QwtScaleDiv.MajorTick)
        medLen = QwtScaleDraw.tickLength(self,QwtScaleDiv.MediumTick)
        minLen = QwtScaleDraw.tickLength(self,QwtScaleDiv.MinorTick)
        #print('majLen medLen ', majLen, ' ', medLen, ' ', minLen)
        #print('lengths ', majLen,' ', medLen,' ', minLen)
        self.offset = 0
        scldiv = QwtScaleDraw.scaleDiv(self)
        # plot major ticks
        # Sometimes, for reasons deep inside Qwt, major_ticks seems to be 
        # returned as a null list for reasons I don't understand
        major_ticks = scldiv.ticks(QwtScaleDiv.MajorTick)
        #print('major ticks ', major_ticks)
        major_step = 0.0
        try:
          major_step = major_ticks[1] - major_ticks[0]
        except:
          pass
        major_count = len(major_ticks)
        for i in range(major_count):
            val = major_ticks[i]
            #print('initial major val = ', val)
            if not self.end_value is None:
              v = val
              if val >= self.end_value:
                v = val - self.delta
            else:
              if not self.divisor is None:
                v = val % self.divisor
            if self.offset == 0 and v != val:
                self.offset = major_step - v % major_step
            val = val + self.offset
            #print('final major val = ', val)
            QwtScaleDraw.drawTick(self, painter, val, minLen)
            QwtScaleDraw.drawLabel(self, painter, val)

        # also, optionally plot a major tick at the dividing point
        if self.do_separator:
          QwtScaleDraw.drawTick(self,painter, self.separator, majLen)
          QwtScaleDraw.drawLabel(self,painter,self.separator)


        # Sometimes, for reasons deep inside Qwt, minor_ticks seems to be 
        # returned as a null list for reasons I don't understand
        minor_ticks = scldiv.ticks(QwtScaleDiv.MinorTick)
        minor_step = 0.0
        try:
          minor_step = minor_ticks[1] - minor_ticks[0]
        except:
          pass
        minor_count = len(minor_ticks)

        # can't handle logs properly, but so what?
#       if scldiv.logScale():
#           pass
#           for i in range(scldiv.minCnt()):
#               QwtScaleDraw.drawTick(
#                   self, painter, scldiv.minMark(i), minLen)
#       else:
        if True:
            minor_ticks = scldiv.ticks(QwtScaleDiv.MinorTick)
            #print('minor ticks ', minor_ticks)
            medium_ticks = scldiv.ticks(QwtScaleDiv.MediumTick)
            #print('medium ticks ', medium_ticks)
            kmax = major_count - 1
            if kmax > 0:
                majTick = major_ticks[0]
                hval = majTick - 0.5 * major_step
                k = 0
                for i in range(minor_count):
                    val = minor_ticks[i]
                    if  val > majTick:
                        if k < kmax:
                            k = k + 1
                            majTick = major_ticks[k]
                        else:
                            majTick += (major_ticks[kmax]
                                        + major_step)
                        hval = majTick - 0.5 * major_step
                    if abs(val-hval) < step_eps * major_step:
                        QwtScaleDraw.drawTick(self,painter, val, medLen)
                    else:
                        QwtScaleDraw.drawTick(self, painter, val, minLen)

        if QwtScaleDraw.Backbone:
            QwtScaleDraw.drawBackbone(self, painter)
    # draw()
# class ComplexScaleDraw()

# stuff for testing
def main(args):
    app = QApplication(args)
    demo = QwtPlot()
    grid = QwtPlotGrid()
    grid.attach(demo)
    grid.setPen(QPen(Qt.black, 0,Qt.DotLine))
    grid.enableX(True)
    grid.enableY(True)
    complex_divider = 50.0

    myXScale = ComplexScaleDraw(start_value=0.0, end_value=complex_divider)
    #print('myXScale', myXScale)
    demo.setAxisScaleDraw(QwtPlot.xBottom, myXScale)

    m = QwtPlotMarker()
    m.attach(demo)
    m.setValue(complex_divider, 0.0)
    m.setLineStyle(QwtPlotMarker.VLine)
    m.setLabelAlignment(Qt.AlignRight | Qt.AlignBottom)
    m.setLinePen(QPen(Qt.black, 2, Qt.SolidLine))

    vector_array = numpy.zeros((100,), numpy.float32)
    for i in range(100):
      vector_array[i] = i

    curve = QwtPlotCurve('example data')
    curve.attach(demo)
    x_array = numpy.zeros(100, numpy.float32)
    y_array = numpy.zeros(100, numpy.float32)
    for i in range(100):
      x_array[i] = 1.0 * i
      y_array[i] = 2.0 * i
    curve.setSamples(x_array,y_array)

    demo.resize(600, 400)
    demo.replot()
    demo.show()
#   app.setMainWidget(demo)
    app.exec_()
# main()

if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from qwt.qt.QtGui import QTreeWidget, QTreeWidgetItem, QApplication, QWidget
from qwt.qt.QtCore import Qt, QObject, pyqtSignal


# some simple classes to create a tree-like structure for viewing and 
# selecting Vells data elements

#===========================================================================

class VellsElement( QTreeWidgetItem ) :
  """
  Inherits from QTreeWidgetItem so that we can store and return keys
  """
  def __init__( self, parent=None, after=None):
    if after is None:
      QTreeWidgetItem.__init__( self, parent)
    else:
      QTreeWidgetItem.__init__( self, parent, [after] )    # ....?
    self.key = None

  def setKey(self, key):
    """
    define a key associated with the QTreeWidgetItem
    """
    self.key = key

  def getKey(self):
    """
    return the key associated with the QTreeWidgetItem
    """
    return self.key


class VellsView(QTreeWidget) :
  """
  inherits from QListView so that we can get keys from VellsElements
  """
  selected_vells_id = pyqtSignal()

  def __init__( self, parent=None, name=None):
    QTreeWidget.__init__( self, parent)
    self.itemClicked[QTreeWidgetItem, int].connect(self.slotVellSelected)

    self.header().hide();

    # self.setSorting(-1)
    # self.setRootIsDecorated( True );

  # get and return the key associated with the element we just clicked
  def slotVellSelected(self,i):
    try:
      result = i.getKey()
      if not result is None:
        self.selected_vells_id.emit(result)
    except:
      pass

def main(args):
  app = QApplication(args)
  m_treeView =  VellsView()
  root = VellsElement( m_treeView, "result" )
  a = VellsElement( root , 'root');
  a.setText(0,"beginning")
  b = VellsElement( root, 'beginning')
  b.setText(0,"item 0")
  b.setKey("Pig")
  e = VellsElement(root, 'pig');
  e.setText(0,"item 5")
  e.setKey("Chicken")
  c = VellsElement( root, 'chicken')
  c.setText(0,"item 10")
  c.setKey("Ape")
  d = VellsElement(a, 'donkey');
  d.setText(0,"item one")
  d.setKey("Donkey")

  rect = QApplication.desktop().geometry();
  m_treeView.move(rect.center() - m_treeView.rect().center())
  m_treeView.show()
  app.exec_()

# Admire
if __name__ == '__main__':
    main(sys.argv)




#!/usr/bin/env python3

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Timba
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *
from Timba.Meq import meqds
from Timba.Meq.meqds import mqs
from Timba.Meq import meq
from MeqGUI.GUI import browsers
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI.treebrowser import NodeAction
import MeqGUI.GUI.app_proxy_gui
from MeqGUI.GUI.meqserver_gui import makeNodeDataItem
from MeqGUI import Grid


from qt import *
import weakref
from qttable import *



_dbg = verbosity(0,name='node_fiddle');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;

solverstart =5;


parmfiddler_instructions = \
                         '''The Parmfiddler shows all the MeqParms as well as all solvers in the tree below the node it is started from.<br> 
                         The funklet/default_funklet/solvability/scales and offsets for the selected Parm(s) can be adjusted via the fiddler. The first coefficient of the funklet of the selected Parm(s) can be easily changed with the slider. The <b>Zero</b> button is a shortcut to reset all coefficients of the selected Parm(s) to 0. <br>To allow for  multiple selection use the <b>multi</b> checkbox, if checked all changes are applied to any selected (highlighted) Parm.  If the <b>reexecute</b> box is checked, the node from which the fiddler is started is reexecuted automatically after each change.<br> Use the right mouse button for extra options.'''





class NA_ParmFiddler(NodeAction):
  text = "ParmFiddler";
  nodeclass = meqds.NodeClass('MeqNode');
  iconset = pixmaps.fiddle.iconset;
  def activate (self,node):
      Grid.Services.addDataItem(makeNodeDataItem(node,viewer=ParmFiddler));

  def is_enabled (self,node):
    # available in idle mode, or when stopped at a debugger. 
    # If debug_level is set, node must be idle (otherwise, we can't trust the
    # node control status to be up-to-date)
    return ( self.tb.is_stopped or not self.tb.is_running ) and \
           ( not self.tb.debug_level or node.is_idle() );


_request_type = dmi_type('MeqRequest',record);




class ParmFiddler (browsers.GriddedPlugin):
  viewer_name = "ParmFiddler";
  _icon = pixmaps.fiddle;


  def is_viewable (data):
    return isinstance(data,_request_type) or \
      isinstance(getattr(data,'request',None),_request_type);
  is_viewable = staticmethod(is_viewable);




  def __init__(self,gw,dataitem,cellspec={},**opts):
    browsers.GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);

    _dprint(1,"started with",dataitem.udi,dataitem.data);
    self._has_data = False;
    self.rid=0;
    self.c00=0.;
    self._wtop = QHBox(self.wparent());
    self._parmlist= [];
    self._solverdict= {};
    self._parmdict= {};
    #all names off the nodes except parms
    self._nodedict= {};
    self._currentparm = None;
    self._request = None;
    self.swapsort=True;
    self.sorted=0; 
    # Initialize subscribe state
    self._callbackparm = curry(self._update_parmlist);
    self._callbacksolver = curry(self._update_solverlist);


    self.parmtable = QTable(self._wtop , "parmtable" );

    QWhatsThis.add(self._wtop,parmfiddler_instructions);



    self.parmtable.setShowGrid(False);
    self.parmtable.setFocusStyle(QTable.FollowStyle);
    self.parmtable.setSelectionMode (QTable.MultiRow );




    self._menu=QPopupMenu(self.parmtable);
    self._menu.insertItem(pixmaps.publish.iconset(),"Publish",self.publish_selected);
    self._menu.insertItem(pixmaps.table.iconset(),"Funklet",self.parmSelected);
    tooltip="invert publishing of all selected rows";
    QToolTip.add(self._menu,tooltip);
   
    tooltip="Double Click to get complete funklet, right mouse for more options";
    QToolTip.add(self.parmtable,tooltip);
    QObject.connect( self.parmtable, SIGNAL("selectionChanged()"), self.parmSetIndex )
    QObject.connect( self.parmtable, SIGNAL("doubleClicked(int, int, int, const QPoint &)"), self.parmSelected )
    QObject.connect( self.parmtable, SIGNAL("clicked(int, int, int, const QPoint &)"), self.parmClicked)
    QObject.connect( self.parmtable, SIGNAL("contextMenuRequested ( int, int, const QPoint &)"), self.rightMouse);
    self.sliderframe = QVBox(self.wtop());
#    self.c00frame = QHBox(self.sliderframe);

    self.labelParm = QLabel("no parm selected", self.sliderframe ,"labelParm");
    self.c00Text = QLabel(self.sliderframe ,"c00text");
    tooltip="Selected Parm";
    QToolTip.add(self.labelParm,tooltip);
    alignment = Qt.AlignLeft|Qt.SingleLine;
    self.c00Text.setAlignment(alignment);
    self.labelParm.setAlignment(alignment);
    style = QFrame.StyledPanel|QFrame.Sunken;
    self.c00Text.setFrameStyle(style);
    
  
    self.slider1 = QSlider (self.sliderframe,"slider1");
    tooltip="Adjust c00 of selected Parm and reexecute";
    QToolTip.add(self.slider1,tooltip);
    self.slider1.setMinValue ( -50 );
    self.slider1.setMaxValue ( 50 );

    self.slider1.setTickInterval(5);
    self.slider1.setTickmarks(QSlider.Both);
    
    self.slider1.setTracking(True);
    self.slider1.setEnabled(False);
    
    QObject.connect(self.slider1,SIGNAL("valueChanged(int)"),self.changeC00Text);
    QObject.connect(self.slider1,SIGNAL("sliderReleased()"),self.updateC00);

    self.spinframe = QHBox(self.sliderframe);
    self.labelSpin =  QLabel("stepsize:", self.spinframe ,"labelSpin");
    self.stepsizeList = QComboBox(False, self.spinframe ,"stepsizeList");
    tooltip="Adjust stepsize of slider";
    QToolTip.add(self.stepsizeList,tooltip);
    QObject.connect(self.stepsizeList,SIGNAL("activated(int)"),self.changeStepsize);
    self.fillstepsizelist();
    self.stepsize=1.;


    self.enable_exec = QCheckBox("reexecute",self.sliderframe, "enable_exec");
    tooltip="If checked, node is executed after change";
    QToolTip.add(self.enable_exec,tooltip);
    self.enable_exec.setEnabled(False);


    self.enable_multi = QCheckBox("multi",self.sliderframe, "multi");
    tooltip="If checked, multiple parms can be adjusted at the same time";
    QToolTip.add(self.enable_multi,tooltip);
    self.enable_multi.setEnabled(True);
    self.enable_multi.setChecked(True);
    QObject.connect(self.enable_multi,SIGNAL("clicked()"),self.toggle_multi);
    
    
    self.ButtonFrame = QHBox(self.sliderframe);
        
    self.buttonOk = QPushButton(self.ButtonFrame,"buttonOk")
    self.buttonOk.setText('Execute');
    tooltip="executes node";
    QToolTip.add(self.buttonOk,tooltip);

    self.buttonReset = QPushButton(self.ButtonFrame,"buttonReset")
    self.buttonReset.setText('Zero');
    tooltip="sets all coefficients of selected Parm to 0";
    QToolTip.add(self.buttonReset,tooltip);
    self.buttonOk.setEnabled(False);
    self.buttonReset.setEnabled(False);
    QObject.connect(self.buttonOk,SIGNAL("clicked()"),self.do_reexecute);
    QObject.connect(self.buttonReset,SIGNAL("clicked()"),self.resetfunklet);

    # form a caption and set contents
    (name,ni) = meqds.parse_node_udi(dataitem.udi);
    caption = "Fiddle from <b>%s</b>" % (name or '#'+str(ni),);
    self._node = ni;
    self._name = name;
    # setup widgets
    self.set_widgets(self.wtop(),caption,icon=self.icon());

    if dataitem.data is not None:
        self.set_data(dataitem);


   

    
    

  def wtop(self):
      return self._wtop;


  def toggle_multi(self):
    if self.enable_multi.state():
      self.parmtable.setSelectionMode (QTable.MultiRow );
    else:
      self.parmtable.setSelectionMode (QTable.SingleRow );
      

  def fillstepsizelist(self):
      for i in range(-10,11):
        if  abs(i) <4:
          stepsize = pow(10.,i);
          self.stepsizeList.insertItem(str(stepsize));
        else:
          sign = "+";
          if i<0:
            sign = "-";
          
          stpstring = "1e"+sign+str(abs(i));
          self.stepsizeList.insertItem(stpstring);
         
      self.stepsizeList.setCurrentItem(10);
      self.stepsize=1;
      
  def changeStepsize(self,stepnr=10):
    self.stepsize= pow(10.,stepnr-10);


  def set_data (self,dataitem,**opts):
      request = getattr(dataitem.data,'request',None);
      if self._request and request:
        self._request  =  request;
      
      
      if not self._has_data:
        self._has_data = True;
        state = dataitem.data;
        if not request:
            QMessageBox.warning(self.wtop(),
                                "Warning",
                              "No request found in Node, Please specify request first via Reexecute");
            self.enable_exec.setOn(False);
            self.enable_exec.setEnabled (False);
            self.buttonOk.setEnabled(False);          
        else:
            #
            #     #print self._node;
            self.enable_exec.setOn(True);
            self.enable_exec.setEnabled (True);
            self.buttonOk.setEnabled(True);          
            self._request = request;

        self.unsubscribe_all();
        #parmlist is sorted by name.
        self._parmlist= [];
        self._solverdict= {};
        self._parmdict= {};
        #all names off the nodes except parms
        self._nodedict= {};


        # recursively get all parms and solvers behind this node
        if self._node:
          self.getnodelist(meqds.nodelist[self._node]);
        else:
          self.getnodelist(meqds.nodelist[self._name]);          
        #subscribe to solvers
        for solverkey in list(self._solverdict.keys()):
          solver = self._solverdict[solverkey]['node'];
          # subscribe
          solver.subscribe_state(self._callbacksolver);
          meqds.request_node_state(solver);

        # fill listbox with MeqParm names
        self.parmtable.setNumCols(len(list(self._solverdict.keys()))+solverstart);
        self.parmtable.setNumRows(len(self._parmlist));
        self.parmtable.setReadOnly(True);
        self.parmtable.horizontalHeader () .setLabel(0,"name");
        self.parmtable.horizontalHeader () .setLabel(1,"c00");
        self.parmtable.horizontalHeader () .setLabel(2,"shape");
        self.parmtable.horizontalHeader () .setLabel(3,pixmaps.publish.iconset(),"");
        self.parmtable.horizontalHeader () .setLabel(4,"0");
        self.parmtable.setColumnWidth(0,100);
        self.parmtable.setColumnWidth(1,200);
        self.parmtable.setColumnWidth(2,60);
        self.parmtable.setColumnWidth(3,25);
        self.parmtable.setColumnWidth(4,25);
#        self.parmtable.horizontalHeader () .setLabel(3,"i");
        tooltip="click to sort";
        QToolTip.add(self.parmtable.horizontalHeader (),tooltip);
        QObject.connect( self.parmtable.horizontalHeader (),SIGNAL("released(int)"),self.sortColumn);
        if self._currentparm:
            self._currentparm.reject();
            self._currentparm=None;
        self._parmindex=-1;

        self.updateTable();
        for parmkey in self._parmlist:
          parm =self._parmdict[parmkey]['node'];
          parm.subscribe_state(self._callbackparm);
          meqds.request_node_state(parm);


        self.buttonReset.setEnabled(False);

        if not self._parmlist:
            QMessageBox.warning(self.wtop(),
                                "Warning",
                                "No parameters found");
            self.buttonReset.setEnabled(False);

        self.enable();
        self.flash_refresh();





  def getnodelist(self,node):
      if node.classname == 'MeqParm':
          self.checkparm(node);
      else:
          # check if we have been at this node before
          if not self.checknodenew(node):
              # improves speed
              return;
          if node.classname == 'MeqSolver':
              self._solverdict[node.name]={'group':'','node':node,'parms':{},'col':-1};

              
          if node.children:
              for (key,ni) in node.children:
                  child = meqds.nodelist[ni];
           
                  self.getnodelist(child);
                  
          
  def checknodenew(self,node):
      if node.name in self._nodedict:
          return False;
      
      self._nodedict[node.name] = 1; 
      return True;   


      

  def checkparm(self,node):
    for i in range(len(self._parmlist)):
      checknode=self._parmlist[i];
      if checknode==node.name:
        return False;
      if(checknode>node.name):
        # store in alphabetic order
        self._parmlist.insert(i,node.name);
        self._parmdict[node.name] = {'node' : node, 'name': node.name,'solvers':{},'c00':0.,'shape':'','publish':0,'groups':[],'zero':0,'row':-1,'funklet':{},'class':'MeqPolc','defaultclass':'MeqPolc'};
        return True;

    self._parmlist.append(node.name);
  
    self._parmdict[node.name] = {'node':node, 'name': node.name,'solvers':{},'c00':0.,'shape':'','publish':0,'groups':[],'zero':0,'row':-1,'funklet':{},'class':'MeqPolc','defaultclass':'MeqPolc'};
      
          
    return True;



  def updateTable(self): 
      #print "got updateTable ",self._parmdict;
      solvernamelist = list(self._solverdict.keys());
      if solvernamelist:
          n=solverstart;
          for solvernm in solvernamelist:
              self.parmtable.horizontalHeader () .setLabel(n,solvernm);              
              self._solverdict[solvernm]['col']=n;
 #             self.parmtable.setColumnReadOnly(n,True);
              n+=1;
      #QObject.connect(self.parmtable.horizontalHeader (),SIGNAL("clicked(int)"),self.sortbycol)
      parmnr=0;
      for parmkey in self._parmlist:
          
          parm =self._parmdict[parmkey]['node'];
          solvers = self._parmdict[parmkey]['solvers'];
          c00 =  self._parmdict[parmkey]['c00'];
          shape =  self._parmdict[parmkey]['shape'];
          publish =  self._parmdict[parmkey]['publish'];
          zero =  self._parmdict[parmkey]['zero'];
          funkletclass = self._parmdict[parmkey]['class'];
          self.parmtable.setText( parmnr,0,parm.name );
          self.parmtable.setText( parmnr,1,str(c00) );
          self.parmtable.setText( parmnr,2,shape );
          if publish:
            self.parmtable.setPixmap(parmnr,3,pixmaps.publish.pm());
          else:
            self.parmtable.clearCell(parmnr,3);
          if zero:
            self.parmtable.setPixmap(parmnr,4,pixmaps.check.pm());
          else:
            self.parmtable.clearCell(parmnr,4);
          self._parmdict[parmkey]['row']=parmnr;
              
          parmnr+=1;

      self.putcheckboxes();


  def sortColumn(self,col):
    swapsort =True;
    if self.sorted==col:
      swapsort = not self.swapsort;
    
    self.parmtable.sortColumn ( col,swapsort,True);
    self.sorted=col;
    self.swapsort = swapsort;
    #update indices
    for i in range(self.parmtable.numRows()):
      parmkey = str(self.parmtable.text(i,0));
      self._parmdict[parmkey]['row']=i;

    # update selected parm
    self._parmindex=-1;
    self.parmSetIndex();
  
    


  def updateRow(self,row):
    parmkey = str(self.parmtable.text(row,0));
    c00=self._parmdict[parmkey]['c00'];
    shape =  self._parmdict[parmkey]['shape'];
    solvers = self._parmdict[parmkey]['solvers'];
    publish =  self._parmdict[parmkey]['publish'];
    zero =  self._parmdict[parmkey]['zero'];
    funkletclass = self._parmdict[parmkey]['class'];
    
#    print "updating parm ",self._parmdict[parmkey];
    self.parmtable.setText( row,1,str(c00) );
    self.parmtable.setText( row,2,shape );
    if publish:
      self.parmtable.setPixmap(row,3,pixmaps.publish.pm());
    else:
      self.parmtable.clearCell(row,3);
    if zero:
      self.parmtable.setPixmap(row,4,pixmaps.check.pm());
    else:
      self.parmtable.clearCell(row,4);
 
    for solver in list(self._solverdict.keys()):
      col = self._solverdict[solver]['col'];
      checkbutton = self.parmtable.item(row,col);
      if not checkbutton:
        return;
      if solver in solvers:
        if solvers[solver]:
          checkbutton.setChecked(True);
     

  def putcheckboxes(self):
     for solver in list(self._solverdict.keys()):
         col = self._solverdict[solver]['col'];
         if 'parms' in self._solverdict[solver]:
             for parm in list(self._solverdict[solver]['parms'].keys()):
                 row =  self._parmdict[parm]['row'];
                 solvers = self._parmdict[parm]['solvers'];
                 checkbutton = QCheckTableItem(self.parmtable, "");
                 checkbutton.setChecked(False);
                 if solver in solvers:
                     if solvers[solver]:
                         checkbutton.setChecked(True);
                 self.parmtable.setItem(row,col,checkbutton);
               


  def _update_parmlist(self,node,state,event=None):
      nodegroups=[];
      try: nodegroups=state.node_groups;
      except AttributeError: pass;
      funklet=None;
      try: funklet=state.funklet;
      except AttributeError: pass;
      default_funklet=None;
      try: default_funklet=state.default_funklet;
      except AttributeError: pass;
      coeff=[[0]];
      funkletclass = 'MeqPolc';
      if funklet:
          coeff=funklet.coeff;
          if hasattr(funklet,'class'):
            funkletclass = getattr(funklet,'class');
      defaultfunkclass = 'MeqPolc';
      if default_funklet:
        if hasattr(default_funklet,'class'):
          defaultfunkclass = getattr(default_funklet,'class');
      if is_scalar(coeff):
          coeff=[[coeff]];
      if is_scalar(coeff[0]):
        ncoeff=[];
        for i in range(len(coeff)):
          ncoeff.append([coeff[i]]);
        coeff=ncoeff;
      zero=1;
      shapex= len(coeff);
      shapey=len(coeff[0]);
      for i in range(shapex):
        for j in range(i==0,shapey): #funny way to exclude 0,0
          
          if not coeff[i][j]==0:
            zero=0;
            break;
        if not zero:
          break
      shapestr="["+str(shapex)+","+str(shapey)+"]";
      self._parmdict[node.name]['funklet']=funklet;
      self._parmdict[node.name]['c00']=coeff[0][0];
      self._parmdict[node.name]['shape']=shapestr;
      self._parmdict[node.name]['publish']=node.is_publishing();
      self._parmdict[node.name]['zero']=zero;
      self._parmdict[node.name]['class']=funkletclass;
      self._parmdict[node.name]['defaultclass']=defaultfunkclass;
      nodegroups=make_hiid_list(nodegroups);
      self._update_subscribedsolvers(node,nodegroups);

  def _update_subscribedsolvers(self,node, nodegroups):
      solvers={};
      for solver in list(self._solverdict.keys()):
          solvers[solver]=0;
          if node.name in self._solverdict[solver]['parms']: #if parm in list of solvable parms, 
              for group in nodegroups:
                  if group==self._solverdict[solver]['group']:
                      solvers[solver]=1;
      self._parmdict[node.name]['solvers']=solvers;
      self._parmdict[node.name]['groups']=nodegroups;

      row = self._parmdict[node.name]['row'];
      self.updateRow(row);      


  def _update_solverlist(self,node,state,event=None):

      names={};
      for i in state.solvable['command_by_list']:
          if 'state' in i and i['state']['solvable']  and 'name' in i:
              nameslist=i['name'];
      if isinstance(nameslist,str):
          names[nameslist]=1;
      else:
          for name in nameslist:
              names[name]=1;
      
      try: nodegroups=state.parm_group;
      except AttributeError: return;
      # get solver from dictionary
      if self._solverdict[node.name]['group'] == nodegroups:
          return;
      else:
         self._solverdict[node.name]['group']=  nodegroups;
         self._solverdict[node.name]['parms']=  names;
         for parm in list(self._parmdict.keys()):
             parmnode=self._parmdict[parm]['node'];
             # update parmtable if needed
             meqds.request_node_state(parmnode);
      self.putcheckboxes();

  def changeC00Text(self, value):
    change=self.c00 + value*self.stepsize;
    self.c00Text.setText(str(change));


  def changeC00(self, value=0.00):
    self.c00=value;
    if(self.slider1.value()==0):
      self.changeC00Text(0);
    self.slider1.setValue(0);
    
  def updateC00(self):
    ok = True;
    change = self.c00Text.text().toDouble()[0];
    self.changeC00(change);

     

    if self.enable_multi.state():
      #change all selected
      for parmkey in self._parmlist:
        #      if not self._parmdict[parmkey]['publish']:
        if self.parmtable.isRowSelected(self._parmdict[parmkey]['row']):
          #set c00
          funklet = self._parmdict[parmkey]['funklet'];
          coeff=array([[0.]]);
          if funklet:
            coeff=funklet.coeff;
          if is_scalar(coeff):
            coeff=array([[self.c00]]);
            #            coeff=[[coeff]];
          else:
            if is_scalar(coeff[0]):
              #            ncoeff=[];
              #            for i in range(len(coeff)):
              #              ncoeff.append([coeff[i]]);
              #            coeff=ncoeff;
              coeff[0]=self.c00;
            else:
              if is_scalar(coeff[0][0]):
                coeff[0][0]=self.c00;

          funklet.coeff=coeff;
          self._parmdict[parmkey]['funklet']=funklet;

          meqds.set_node_state(parmkey,funklet=funklet);
      self.reexecute();
    else:
      if not self._currentparm:
        return;
      self._currentparm.setc00(self.c00);
      
 
    


  def resetfunklet(self):
      if self.enable_multi.state():
        #change all selected
        for parmkey in self._parmlist:
          #      if not self._parmdict[parmkey]['publish']:
          if self.parmtable.isRowSelected(self._parmdict[parmkey]['row']):
            #set to 0
            funklet = self._parmdict[parmkey]['funklet'];
            coeff=array([[0.]]);
            if funklet:
              coeff=funklet.coeff;
            if is_scalar(coeff):
              coeff=array([[0.]]);
            else:
              if is_scalar(coeff[0]):
                for i in range(len(coeff)):
                  coeff[i]=0.;
              else:
                for i in range(len(coeff)):
                  for j in  range(len(coeff[0])):

                    coeff[i][j]=0.;
            funklet.coeff=coeff;
            self._parmdict[parmkey]['funklet']=funklet;
#            print "setting funklet ",funklet;
            meqds.set_node_state(parmkey,funklet=funklet);
        self.reexecute();

      else:
        if not self._currentparm:
          return;
        self._currentparm.resettozero();



  def parmSetIndex(self ):
      row = self.parmtable.currentRow();
      col = self.parmtable.currentColumn();
     # if col >=solverstart:
      #   return;


      if row>=0 and row <len(self._parmlist) and not row == self._parmindex :
          self._parmindex = row;
      else:
          #nothing
          return;

      self.buttonOk.setEnabled(True);
      self.buttonReset.setEnabled(True);
      self.slider1.setEnabled(True);
      changenodekey=str(self.parmtable.text(row,0));
      changenode=self._parmdict[changenodekey]['node'];
      if self._currentparm:
          self._currentparm.reject();
      #          del self._currentparm;
      self._currentparm = ParmChange(self,changenode);

      label="c00 of "+changenodekey;
      self.labelParm.setText(label);

      
  def parmSelected(self,row=0,col=0,button=0,mousePos=0):
    if col>=solverstart :
      #nothing
      return;

    row = self.parmtable.currentRow();
    if row>=0 and row <len(self._parmlist):
      self._parmindex = row;
    else:
      #nothing
      return;
    self.buttonOk.setEnabled(True);
    self.buttonReset.setEnabled(True);
    self.slider1.setEnabled(True);
    changenodekey=str(self.parmtable.text(row,0));
    changenode=self._parmdict[changenodekey]['node'];
    if self._currentparm:
      self._currentparm.reject();
    #          del self._currentparm;
    self._currentparm = ParmChange(self,changenode);


    self.getparms();


      
  def parmClicked(self,row,col,button,mousePos):
  
    if button == Qt.LeftButton:
      self.leftMouse(row,col);
    if button == Qt.RightButton:
      self.rightMouse(row,col,mousePos);
    if button == Qt.MidButton:
      self.rightMouse(row,col,mousePos);
      

    
  def leftMouse(self,row,col):
    #check if solvable paramters are updated
    if row<0 or row >len(self._parmlist) or col > len(list(self._solverdict.keys()))+solverstart:
      #nothing
      return;
    if col <solverstart :
      if not row == self._parmindex:
        self.parmSetIndex();
      #nothing
      return;
    checkbutton = self.parmtable.item(row,col);
    if not checkbutton:
        return;
    parmkey = str(self.parmtable.text(row,0));
    parm=self._parmdict[parmkey]['node'];    
    solver = str(self.parmtable.horizontalHeader().label(col));
    solvers = self._parmdict[parmkey]['solvers'];
    if checkbutton.isChecked():
      checkbutton.setChecked(False);
    else:
      checkbutton.setChecked(True);
      
    updateneeded = False;
    if solver in solvers:
        if solvers[solver]  and not checkbutton.isChecked():
          solvers[solver]=0;
          self._parmdict[parmkey]['solvers']=solvers;
          updateneeded=True;
        if not solvers[solver]  and checkbutton.isChecked():
          solvers[solver]=1;
          self._parmdict[parmkey]['solvers']=solvers;
          updateneeded=True;
        
          
    if updateneeded:
      nodegroups=[];
      for solver in list(self._solverdict.keys()):
        group = self._solverdict[solver]['group'];
        if solver in solvers:
          if solvers[solver]:
            nodegroups.append(group);
      dmigroups=make_hiid_list(nodegroups);
      if nodegroups:
        meqds.set_node_state(parm,node_groups=dmigroups);
      else:
        meqds.set_node_state(parm,node_groups=hiid(0));
        meqds.set_node_state(parm,solvable=False);
      
  def rightMouse(self,row,col,mousePos):
    self._menu.popup(mousePos);
       
  def publish_selected(self):
    #get selected rows
    for parmkey in self._parmlist:
#      if not self._parmdict[parmkey]['publish']:
        if self.parmtable.isRowSelected(self._parmdict[parmkey]['row']):
          meqds.enable_node_publish(parmkey,enable=not self._parmdict[parmkey]['publish']);


     
  
  def getparms(self):
    if not self._parmlist:
        QMessageBox.warning(self.wtop(),
                            "Warning",
                            "No parameters found");
        return;

    if self._parmindex >= len(self._parmlist) or self._parmindex< 0 :
        QMessageBox.warning(self.wtop(),
                            "Warning",
                            "No parameter selected");
        return;




    changenodekey=str(self.parmtable.text(self._parmindex,0));
    changenode=self._parmdict[changenodekey]['node'];
    if not self._currentparm:
      self._currentparm = ParmChange(self,changenode);
    if self._currentparm.edit_parm :
      self._currentparm.rejectedit();
    # create new editor
    self._currentparm._edit();



  def update_default_selected(self,funklet):
    newfunkclass = 'MeqPolc';
    if hasattr(funklet,'class'):
      newfunkclass=getattr(funklet,'class');
    for parmkey in self._parmlist:
      if self.parmtable.isRowSelected(self._parmdict[parmkey]['row']):
        funkletclass = self._parmdict[parmkey]['defaultclass'];
        if not newfunkclass==funkletclass:
          # cannot change class
          continue;
#        print "setting def_funklet ",funklet;
        meqds.set_node_state(parmkey,default_funklet=funklet);    


  def update_selected(self,funklet):
    newfunkclass = 'MeqPolc';
    if hasattr(funklet,'class'):
      newfunkclass=getattr(funklet,'class');
    for parmkey in self._parmlist:
      if self.parmtable.isRowSelected(self._parmdict[parmkey]['row']):
        funkletclass = self._parmdict[parmkey]['class'];
        if not newfunkclass==funkletclass:
          # cannot change class
          continue;
        self._parmdict[parmkey]['funklet']=funklet;
#        print "setting funklet ",funklet;
        meqds.set_node_state(parmkey,funklet=funklet);
 



  def _refresh (self):
    # clear the has_data flag, so that set_data above will accept new data
    _dprint(2,'refresh expected now');
    self._has_data = False;

  def do_reexecute(self):
      self.reexecute(True);

  def reexecute(self,do_execute = False):

    if not self.enable_exec.state () and not do_execute: #reexecute not enabled
      #print "auto execute not eneabled"
      return;
    
    if not self._request:
      _dprint(2,'no request to execute');
      return;
    #check solver status:
    #      self.updateTable();
    _dprint(1,'accepted: ',self._request);
    reqid_old= self._request.request_id;
    if not reqid_old:
      reqid_old=  hiid(1,1,1,1);

    rid0=int(reqid_old[0]);
    rid1=int(reqid_old[1])+1;
    rid2=int(reqid_old[2])+1;
    reqid=hiid(rid0,rid1,rid2,0);


    self._request.request_id = reqid;
    #self._request.cache_override = True;
    if self._node:
      cmd = record(nodeindex=self._node,request=self._request,get_state=True);
    elif self._name:
      cmd = record(name=self._name,request=self._request,get_state=True);
    mqs().meq('Node.Execute',cmd,wait=False);



  def unsubscribe_all(self):
      for parmkey in list(self._parmdict.keys()):
          parm =  self._parmdict[parmkey]['node'];
          #print "unsubscribe node ", parm.name;
          parm.unsubscribe_state(self._callbackparm);
      #print "unsubscribing solvers", self._solverdict;
      for solverkey in list(self._solverdict.keys()):
          solver = self._solverdict[solverkey]['node'];
          #print "unsubscribe solver ", solver.name;
          solver.unsubscribe_state(self._callbacksolver);

  def cleanup (self):
      browsers.GriddedPlugin.cleanup(self);
      
      #unsubscribe
      
      self.unsubscribe_all();

      if self._currentparm:
          self._currentparm.reject();
          del self._currentparm;





class ParmChange:
  def __init__(self,parent,node):
      self._parent=parent;
      self._node=node;
      self._nodename=node.name;      
#      self._callback = curry(self._update_state);
#      self._node.subscribe_state(self._callback);
#      meqds.request_node_state(self._node);
      self.edit_parm=None;
      self._funklet=None;
      self._dorefresh();

  def _dorefresh (self):
    # request node state, and subscribe to it via the curried callback
    # the curry() is handy because it will automatically disconnect the
    # subscription when deleted
    self._callback = curry(self._update_state);
    self._node.subscribe_state(self._callback);
    meqds.request_node_state(self._node);
      
  def _update_state(self,node,state,event=None):
      try: funklet = state.funklet;
      except AttributeError: #pass
          #      QListViewItem(self.reqView.wlistview(),'','','(no funklet found in node)');
          
          #pop up
          #        self.buttonOk.setEnabled(False);
          #          print "no funklet found in node";
          QMessageBox.warning(None,
                              "Warning",
                              "No funklet found in Node");

      else:
          #print funklet;
          #      print self._node;
          #      self.buttonOk.setEnabled(True);
          self._funklet = state.funklet;
          self._parent.changeC00(self.getc00());
          if self.edit_parm:
              self.edit_parm.updateCoeff_fromparent();

  def _edit(self):
      if self.edit_parm:
          self.reject();
      self.edit_parm=editParm(self,self._parent.wtop());

  def getc00(self):
      if not self._funklet:
          return 0.;
      coeff = self._funklet.coeff;
      if is_scalar(coeff):
          return self._funklet.coeff;
      else:
          if is_scalar(coeff[0]):
              return self._funklet.coeff[0];
      return self._funklet.coeff[0][0];

  def setc00(self,value):
      if not self._funklet:
          return 0.;
      if is_scalar(self._funklet.coeff):
          self._funklet.coeff=value;
      else:
          if is_scalar(self._funklet.coeff[0]):
              self._funklet.coeff[0]=value;
          else:
              self._funklet.coeff[0][0]=value;
      self.updatechange();
      if self.edit_parm:
          self.edit_parm.updateCoeff_fromparent();



  def resettozero(self):
      if not self._funklet:
        return 0;
      if is_scalar(self._funklet.coeff):
          self._funklet.coeff=array([[0.]]);
      else:
          if is_scalar(self._funklet.coeff[0]):
              for i in range(len(self._funklet.coeff)):
                  self._funklet.coeff[i]=0.;
          else:
              for i in range(len(self._funklet.coeff)):
                  for j in range(len(self._funklet.coeff[i])):
                                 self._funklet.coeff[i][j]=0.;
      self.updatechange();
      if self.edit_parm:
          self.edit_parm.updateCoeff_fromparent();




  def setDefault(self):
      if not self._funklet:
         return;
      if not self._parent:
          self.reject();
      self._parent.update_default_selected(self._funklet);

#      meqds.set_node_state(self._node,default_funklet=self._funklet);


  def updatechange (self):
      if not self._funklet:
         return;
      if not self._parent:
          self.reject();


      self._parent.update_selected(self._funklet);

      self._parent.changeC00(self.getc00());
      
      self._parent.reexecute();

  def rejectedit(self):
       if self.edit_parm:
          self.edit_parm.reject();
          self.edit_parm=None;
     

  def reject(self):
      self._node = self._callback = None; # this will disconnect the Qt signal
      self.rejectedit();


class editParm(QDialog):
    def __init__(self, parent,grandparent):
      self.parent = parent;
      self.grandparent = grandparent;
      self.parentname=parent._nodename;
      funklet=self.parent._funklet;
      if funklet:
          self._coeff=funklet.coeff;
      else:
          self._coeff=[[0.]];

      if is_scalar(self._coeff):
          self._coeff=[[self._coeff]]
      if is_scalar(self._coeff[0]):
        ncoeff=[];
        for i in range(len(coeff)):
          ncoeff.append([coeff[i]]);
        coeff=ncoeff;
      self._nx=len(self._coeff);
      self._ny=len(self._coeff[0]);

      QDialog.__init__(self,grandparent,"Test",0,0);
      self.setCaption('Funklet of '+self.parentname);
      self.v = QVBoxLayout(self, 10, 5)

      self.funkgrid=QTable(self);
      self.funkgrid.setCaption("funklet values")

      self.horh = self.funkgrid.horizontalHeader()
      self.verth = self.funkgrid.verticalHeader()

#      self.updateCoeff_fromparent();

      

      self.v.addWidget(self.funkgrid);


      self.Bh = QHBoxLayout(self.v, 5)

      
      self.cmdAddRow = QPushButton('Add Row', self)
      QObject.connect(self.cmdAddRow, SIGNAL('clicked()'), self.slotcmdAddRow)
      self.Bh.addWidget(self.cmdAddRow)
      self.cmdRemoveRow = QPushButton('Remove Row', self)
      QObject.connect(self.cmdRemoveRow, SIGNAL('clicked()'), self.slotcmdRemoveRow)
      self.Bh.addWidget(self.cmdRemoveRow)
      self.cmdAddCol = QPushButton('Add Column', self)
      QObject.connect(self.cmdAddCol, SIGNAL('clicked()'), self.slotcmdAddCol)
      self.Bh.addWidget(self.cmdAddCol)
      self.cmdRemoveCol = QPushButton('Remove Column', self)
      QObject.connect(self.cmdRemoveCol, SIGNAL('clicked()'), self.slotcmdRemoveCol)
      self.Bh.addWidget(self.cmdRemoveCol)
      self.Bh0 = QHBoxLayout(self.v, 5)
      self.offsLabel = QLabel("Offset (t,f)",self);
      self.Bh0.addWidget(self.offsLabel);
      self.offsF = QLineEdit(self);
#      QObject.connect(self.offsF, SIGNAL(' textChanged(QString)'), self.slotcmdUpdateOffs)
      self.Bh0.addWidget(self.offsF);
      self.offsT = QLineEdit(self);
#      QObject.connect(self.offsT, SIGNAL(' textChanged(QString)'), self.slotcmdUpdateOffs)
      self.Bh0.addWidget(self.offsT);

      self.Bh1 = QHBoxLayout(self.v, 5)
      self.scaleLabel = QLabel("Scale (t,f)",self);
      self.Bh1.addWidget(self.scaleLabel);
      self.scaleF = QLineEdit(self);
      self.Bh1.addWidget(self.scaleF);
      self.scaleT = QLineEdit(self);
      self.Bh1.addWidget(self.scaleT);

      self._offset=array([0.,0.]);
      self._scale=array([1.,1.]);

      self.Bh2 = QHBoxLayout(self.v, 5)
      
      self.cmdOK = QPushButton('Apply', self)
      QObject.connect(self.cmdOK, SIGNAL('clicked()'), self.slotcmdOK)

      self.cmdOK.setFocus();
      
      self.Bh2.addWidget(self.cmdOK)
      self.cmdCancel = QPushButton('Cancel', self)
      QObject.connect(self.cmdCancel, SIGNAL('clicked()'), self.slotcmdCancel)
      self.Bh2.addWidget(self.cmdCancel);
      self.cmdDefault = QPushButton('Set as Default', self)
      QObject.connect(self.cmdDefault, SIGNAL('clicked()'), self.slotcmdDefault)
      self.Bh2.addWidget(self.cmdDefault)

      self.cmdCancel.setIconSet(pixmaps.cancel.iconset());
      self.cmdOK.setIconSet(pixmaps.check.iconset());

      QObject.connect(self.funkgrid,SIGNAL('valueChanged(int,int)'),self.updateCoeff);
      
      self.updategrid();

            
      
      self.show()
      
    def UpdateOffs(self):
#      print "updating offset ",float(str(self.offsF.text()));
      self._offset[0]=float(str(self.offsF.text()));
      self._offset[1]=float(str(self.offsT.text()));
 
    def UpdateScale(self):
#      print "updating scale ",self._scale;
      self._scale[0]=float(str(self.scaleF.text()));
      self._scale[1]=float(str(self.scaleT.text()));
      
    def slotcmdAddRow(self):
        array1=self._coeff;
        array2=Timba.array.resize(array1,(self._nx+1,self._ny));
#       something I have to do , since increasing size of a Timba.array does strange things with values
        for i in range(self._nx):
            for j in range(self._ny):
                array2[i,j]=self._coeff[i][j];
        self._coeff=array2;
        for i in range(self._ny):
            self._coeff[self._nx][i]=0.0;
        self._nx+=1;
        self.funkgrid.setNumRows(self._nx);
        self.funkgrid.setNumCols(self._ny);
        self.updategrid();



    def slotcmdAddCol(self):
        array1=self._coeff;
        array2=Timba.array.resize(array1,(self._nx,self._ny+1));
        for i in range(self._nx):
            for j in range(self._ny):
                array2[i,j]=self._coeff[i][j];
        self._coeff=array2;
        for i in range(self._nx):
            self._coeff[i][self._ny]=0.0;
        self._ny+=1;
        self.funkgrid.setNumRows(self._nx);
        self.funkgrid.setNumCols(self._ny);
        self.updategrid();

    def slotcmdRemoveRow(self):
        if self._nx<=1:
            return;
        array1=self._coeff;
        array2=Timba.array.resize(array1,(self._nx-1,self._ny));
        self._coeff=array2;
        self._nx-=1;
        self.funkgrid.setNumRows(self._nx);
        self.funkgrid.setNumCols(self._ny);
        self.updategrid();
       
        

    def slotcmdRemoveCol(self):
        if self._ny<=1:
            return;
        array1=self._coeff;
        array2=Timba.array.resize(array1,(self._nx,self._ny-1));
        self._coeff=array2;
        self._ny-=1;
        self.funkgrid.setNumRows(self._nx);
        self.funkgrid.setNumCols(self._ny);
        self.updategrid();

    def slotcmdCancel(self):
#        self.parent._dorefresh();
#        self.reject();
        self.parent.rejectedit();


    def slotcmdDefault(self):
      self.slotcmdOK(1);

    def slotcmdOK(self, set_default=0):
#       #print self.parent._funklet.coeff;
        array1=Timba.array.resize(self.parent._funklet.coeff,(self._nx,self._ny));
        self.parent._funklet.coeff=array1;
        for i in range(self._nx):
            for j in range(self._ny):
                
                self.parent._funklet.coeff[i][j] = float(str(self.funkgrid.text(i,j)));

        self.UpdateOffs();
        self.UpdateScale();

        offset=self._offset;
        scale = self._scale;

        self.parent._funklet.offset=offset;
        self.parent._funklet.scale=scale;
        if set_default:
          self.parent.setDefault();
        
#       #print self.parent._funklet.coeff;
#       self.updategrid();
        self.parent.updatechange();
#        self.reject();
                
                
    def updategrid(self):
        self.funkgrid.setNumRows(self._nx);
        self.funkgrid.setNumCols(self._ny);
        maxc=1;
        for j in range(self._ny):
            self.horh.setLabel(j,"");
            for i in range(self._nx):
                if j==0 : self.verth.setLabel(i,""); 
                self.funkgrid.setText(i,j,str(self._coeff[i][j]));
                width= len(self.funkgrid.text(i,j));
                if width>maxc : maxc =width;
            self.funkgrid.setColumnWidth(j,maxc*15);
            maxc=1;
        self.horh.setLabel(0,"f");
        self.verth.setLabel(0,"t");
        self.offsF.setText(str(self._offset[0]));
        self.offsT.setText(str(self._offset[1]));
        self.scaleF.setText(str(self._scale[0]));
        self.scaleT.setText(str(self._scale[1]));

 
    def updateCoeff(self,row,col):
        self._coeff[row][col]= float(str(self.funkgrid.text(row,col)));

    def updateCoeff_fromparent(self):
        funklet=self.parent._funklet;
        self._offset=array([0.,0.]);
        self._scale=array([1.,1.]);
        self._coeff=0.;

        if funklet:
#          print funklet;
          if hasattr(funklet,'coeff'):
            self._coeff=funklet.coeff;
          if hasattr(funklet,'offset'):
            self._offset=funklet.offset;
          if hasattr(funklet,'scale'):
            self._scale = funklet.scale;
        if is_scalar(self._coeff):
            self._coeff=[[self._coeff]]
        if is_scalar(self._coeff[0]):
          ncoeff=[];
          for i in range(len(self._coeff)):
            ncoeff.append([self._coeff[i]]);
          self._coeff=ncoeff;
        self._nx=len(self._coeff);
        self._ny=len(self._coeff[0]);
        self.updategrid();


def define_treebrowser_actions (tb):
  _dprint(1,'defining parm fiddling treebrowser actions');
  tb.add_action(NA_ParmFiddler,30,where="node");

Grid.Services.registerViewer(meqds.NodeClass(),ParmFiddler,priority=25);



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#% $Id$ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
import numpy
HAS_TIMBA = False
try:
  import MeqGUI.Plugins.plotting_functions_qt5 as plot_func
  from Timba.utils import verbosity
  _dbg = verbosity(0,name='VellsData');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
  HAS_TIMBA = True
except:
  pass

class VellsData:
   """ A class for handling and extracting Vells data for display """

   def __init__(self):
     self._active_plane = 0
     self._active_perturb = None

     self._plot_vells_dict = {}
     self._plot_flags_dict = {}
     self._plot_labels = {}
     self._key_menu_labels = {}
     self._menu_labels = {}
     self._menu_labels_big = {}
     self._perturbations_index = {}
     self._planes_index = []
     self.start_vells_id = 499
     self.array_selector = []
     self.array_tuple = None
     self.data_max = None
     self.data_min = None
     self.first_axis_parm = None
     self.second_axis_parm = None
     self.third_axis_parm = None
     self.initialSelection = False
     self.display_3D = False
     self.shape_change = True
     self.scalar_data = False
     self._request_full_image = False
     self._exterior_plot_label = None
     self.rank = -1
     self.actual_rank = -1
     self.shape = (-1,)

#    self.__init__

   def calc_vells_ranges(self, vells_rec):
      """ this method determines the range of a vells. The results
          are used to set up proper axis labelling etc for the
          visualization displays.
      """
      self.axis_labels = []
      self.vells_axis_parms = {}
      self.axis_shape = {}
      self.num_possible_ND_axes = 0
      try:
        axis_map = vells_rec.cells.domain.get('axis_map',['time','freq'])
        if HAS_TIMBA:
          _dprint(3, 'axis map is ', axis_map)
        for i in range(len(axis_map)):
          # convert from Hiid to string
          current_label = str(axis_map[i]).lower()
          if HAS_TIMBA:
            _dprint(3,' ')
            _dprint(3,'current label ', current_label)
          if current_label != '(null)':
            begin = 0
            end = 1
            grid_array = None
            delta = None
            expected_num_grid_points = None
            title = current_label
            if current_label in vells_rec.cells.grid:
              try:
                grid_array = vells_rec.cells.grid.get(current_label)
                delta = vells_rec.cells.cell_size.get(current_label)
                try:
                  grid_step = 0.5 * delta[0]
                except:
                  grid_step = 0.5 * delta
                self.axis_shape[current_label] = grid_array.shape[0]
                if HAS_TIMBA:
                  _dprint(3, 'in calc_vells_ranges: grid_array shape is ', grid_array.shape)
                self.num_possible_ND_axes = self.num_possible_ND_axes + 1
                if HAS_TIMBA:
                  _dprint(3, 'in calc_vells_ranges: incrementing ND axes to ', self.num_possible_ND_axes)
                begin = grid_array[0] - grid_step
                if self.axis_shape[current_label] > 1:
                  grid_step = 0.5 * delta[self.axis_shape[current_label] -1]
                  end = grid_array[self.axis_shape[current_label] -1] + grid_step
                else:
                  end = grid_array[0] + grid_step
                # figure out how many grid points we should have - for now assume that all
                # deltas are same
                expected_num_grid_points = int((end - begin) / delta[0]) 
              except:
                self.axis_shape[current_label] = 1
              title = current_label
              if current_label == 'time':
                end = end - begin
                begin = 0
                title = 'Relative Time(sec)'
              if current_label == 'freq':
                if end >  1.0e6:
                  begin = begin / 1.0e6
                  end = end / 1.0e6
                  title = 'Frequency(MHz)'
                elif end >  1.0e3:
                  begin = begin / 1.0e3
                  end = end / 1.0e3
                  title = 'Frequency(KHz)'
                else:
                  title = 'Frequency(Hz)'
            else:
              self.axis_shape[current_label] = 1
            if HAS_TIMBA:
             _dprint(3,'assigning self.vells_axis_parms key ', current_label)
             _dprint(3,'assigning begin ', begin)
             _dprint(3,'assigning end ', end)
             _dprint(3,'assigning title ', title)
            self.vells_axis_parms[current_label] = (begin, end, title, self.axis_shape[current_label], grid_array, delta, expected_num_grid_points)
            self.axis_labels.append(current_label)
      except:
      # we have no 'cells' field so need to create a fake one for
      # display purposes
        axis_map = ['time','freq']
        self.scalar_data = True
        for i in range(len(axis_map)):
          current_label = axis_map[i]
          begin = 0
          end = 0
          title = current_label
          if current_label == 'time':
            title = 'Relative Time(sec)'
          if current_label == 'freq':
            title = 'Frequency(MHz)'
          self.axis_shape[current_label] = 1
          self.vells_axis_parms[current_label] = (begin, end, title, self.axis_shape[current_label], None, None, None)
          self.axis_labels.append(current_label)

      # do we request a ND GUI?
      if len(self.vells_axis_parms) > 2 and self.num_possible_ND_axes > 2:
        if HAS_TIMBA:
          _dprint(3, '** in calc_vells_ranges:')
          _dprint(3, 'I think I need a ND GUI as number of valid plot axes is ',self.num_possible_ND_axes)
          _dprint(3, 'length of self.vells_axis_parms is ', len(self.vells_axis_parms))
          _dprint(3, 'self.vells_axis_parms is ', self.vells_axis_parms)
          _dprint(3, 'self.axis_labels is ', self.axis_labels)


    # calc_vells_ranges

   def getVellsDataParms(self):
     """ returns vells parameters for use with the visualization display """
     if HAS_TIMBA:
        _dprint(3,'received method call')
     return [self.vells_axis_parms, self.axis_labels, self.num_possible_ND_axes,self.shape]

   def set_exterior_plot_label(self, exterior_plot_label):
     self._exterior_plot_label = exterior_plot_label

   def StoreVellsData(self, vells_rec, rq_label = ''):
     """ converts vells record structure into a format that is
         easy to use with the visualization displays
     """
     self._plot_vells_dict = {}
     self._plot_flags_dict = {}
     self._plot_labels = {}
     self._menu_labels = {}
     self._menu_labels_big = {}
     self._perturbations_index = {}
     self._planes_index = []
     self.scalar_data = False
     self.scalar_string = ""
     self.rq_label = rq_label

     if HAS_TIMBA:
       _dprint(3,' self.rq_label = ', self.rq_label)
     self.calc_vells_ranges(vells_rec)
     if HAS_TIMBA:
       _dprint(3,'now after calc_vells_ranges')
       _dprint(3,'self.scalar_data ', self.scalar_data)
     if self.scalar_data and len(self.rq_label) > 0:
       self.scalar_string = self.rq_label + "\n"

     try:
       self._number_of_planes = len(vells_rec["vellsets"])
     except:
       print('*** failure to get number of planes')
       if len(self.rq_label) > 0:
         self.scalar_string = self.rq_label + ' has no data'
       else:
         self.scalar_string = 'has no data'
       return
     _dprint(3, 'number of planes ', self._number_of_planes)
     id = self.start_vells_id
     self.dims = None
     self.index = []
     self.start_index = None
     if "dims" in vells_rec:
       dims = vells_rec.dims
       self.dims = list(dims)
       for i in range(len(self.dims)):
         self.index.append(0)
       self.start_index = len(self.dims) - 1

# store data
     for i in range(self._number_of_planes):
       try:
         exterior_label = self._exterior_plot_label[i]
       except:
         exterior_label = None
       valueshapes = [];  # accumulate list of main value + perturbed value shapes here
       if "value" in vells_rec.vellsets[i]:
         menu_label = "[" + str(i) + "]" 
         if self.dims is None:
           text_display = menu_label
         else:
           text_display = str(self.index)
         id = id + 1
         self._menu_labels[id] = text_display
         self._key_menu_labels[id] = menu_label
         self._planes_index.append(id)
         self._plot_vells_dict[menu_label] = vells_rec.vellsets[i].value
         # check if we may have to generate enlarged data set
         shape = self._plot_vells_dict[menu_label].shape
         valueshapes.append(shape);
         self._menu_labels_big[id] = False
         for k in range(len(shape)):
           try: 
             expected_size = self.vells_axis_parms[self.axis_labels[k]][6]
           except:
             expected_size = -1
           if expected_size >= 0:
             if self.axis_labels[k].find('freq') >= 0 and expected_size > shape[k]:
               self._menu_labels_big[id] = True
             else:
               self._menu_labels_big[id] = False
         
         if not exterior_label is None:
           tag = "] " + exterior_label + " " + "main value "
         else:
           tag = "] main value "
         if self._number_of_planes > 1:
           if self.dims is None:
             plot_string = "[" + str(i) + tag 
           else:
             plot_string = str(self.index)
         else:
           if self.dims is None:
             plot_string = tag[2:len(tag)] 
           else:
             plot_string = str(self.index)
         if self.scalar_data:
           self.scalar_string = self.scalar_string + plot_string + " " + str(self._plot_vells_dict[menu_label]) + "\n"
         if len(self.rq_label) > 0:
           plot_string = plot_string + " " + self.rq_label
         self._plot_labels[menu_label] = plot_string
       else:
         # there's no actual data - this can happen of one of the vellsets
         # is an empty dictionary, e.g .if we select only diagonal terms
         # of a MS, off-diagonals are supplied as empty dicts.
         # Probably should just get rid of this 'else' - we'll see
         if len(self.rq_label) > 0:
           self.scalar_string = self.rq_label + ' has no data'
         else:
           self.scalar_string = 'has no data'
#        return
       
       if "perturbed_value" in vells_rec.vellsets[i]:
         try:
           number_of_perturbed_arrays = len(vells_rec.vellsets[i].perturbed_value)
           if not exterior_label is None:
             tag = "] " + exterior_label + " " + "perturbed value "
           else:
             tag = "] perturbed value "
           perturbations_list = []
           perturbations_key = str(id) + ' perturbations'
           for j in range(number_of_perturbed_arrays):
             menu_label =  "[" + str(i) + tag + str(j) 
             if self.dims is None:
               text_display = menu_label
             else:
               text_display = str(self.index) + " perturbed value " + str(j)
             id = id + 1
             perturbations_list.append(id)
             self._menu_labels[id] = text_display
             self._key_menu_labels[id] = menu_label
             self._plot_vells_dict[menu_label] = vells_rec.vellsets[i].perturbed_value[j]
             shape = self._plot_vells_dict[menu_label].shape
             valueshapes.append(shape);
             self._menu_labels_big[id] = False
             for k in range(len(shape)):
              expected_size = self.vells_axis_parms[self.axis_labels[k]][6]
              if self.axis_labels[k].find('freq') >= 0 and expected_size > shape[k]:
                self._menu_labels_big[id] = True
              else:
                self._menu_labels_big[id] = False
             if self._number_of_planes > 1:
               initial_plot_str = "[" + str(i) + tag + str(j)
             else:
               initial_plot_str = tag[2:len(tag)] + str(j)
       
             if len(self.rq_label) > 0:
               plot_string = initial_plot_str + " " + self.rq_label
             else:
               plot_string = initial_plot_str
             if self.scalar_data:
               self.scalar_string = self.scalar_string + plot_string + " " + str(self._plot_vells_dict[menu_label]) + "\n"
             self._plot_labels[menu_label] = plot_string
           self._perturbations_index[perturbations_key] = perturbations_list

         except:
           if HAS_TIMBA:
             _dprint(3, 'The perturbed values cannot be displayed.')
       if "flags" in vells_rec.vellsets[i]:
         toggle_index = "flag data " + str(i)
         flags = vells_rec.vellsets[i].flags;
         # OK, we need to promote the shape of the flags array to the union of all the value shapes
         # (usually, all the value shapes are the same anyway, but the flags array can have collapsed axes)
         if valueshapes:
           # figure out the uniting shape
           maxdim = max([len(sh) for sh in valueshapes]);
           maxshape = [ max([(sh[i] if i<len(sh) else 1) for sh in valueshapes]) for i in range(maxdim) ];
           # only do this if uniting shape is non-trivial
           if len(maxshape) > 1 or maxshape[0] > 1:
            # figure out how to reshape flags array to this. We do as follows for e.g. a maxshape of (L,M,N) and a flags shape of (L,1,N)
            # * reshape flags to f0shape, which is a shape containing all non-trivial axes, thus (L,N)
            # * make new flags array of shape L,M,N
            # * assign new[:,:,:] = old[:,.numpy.newaxis,:]
            f0shape = [];
            # build up the indexing object for old (f0index) and new (f1index) arrays
            f0index = [];
            f1index = [numpy.s_[:]]*maxdim;   # this is just :,:,:,...
            # loop over all dimensions
            for idim,dimlen in enumerate(maxshape):
              # if this dimension is missing in flags, add numpy.newaxis to its index
              if idim >= len(flags.shape) or flags.shape[idim] == 1:
                f0index.append(numpy.newaxis);
              # if this dimension is present in flags, add : to its index, and its length to the f0shape
              elif flags.shape[idim] == dimlen:
                f0index.append(numpy.s_[:]);
                f0shape.append(dimlen);
              # else bomb out -- this shouldn't happen, and we won't display such flags
              else:
                flags = None;
                break;
            # now reshape and promote
            if flags is not None:
              flags0 = flags.reshape(f0shape);
              flags = numpy.zeros(maxshape,int);
              if not numpy.isscalar(flags):
                flags[f1index] = flags0[f0index] if not numpy.isscalar(flags0) else flags0;
         if flags is not None:
          self._plot_flags_dict[toggle_index] = flags;

# update index used for strings on displays if self.dims exists
       if not self.dims is None:
         for j in range(self.start_index,-1,-1):
           if self.index[j] < self.dims[j] - 1:
             self.index[j] = self.index[j] + 1
             if j < self.start_index:
               for k in range(j+1, len(self.index)):
                 self.index[k] = 0
             break
           else:
             pass

# initialize axis selection ?
     if not self.initialSelection:
       tag = self._key_menu_labels[self.start_vells_id + 1]
       data = self._plot_vells_dict[tag]
       rank = data.ndim
       shape = data.shape
       self.setInitialSelectedAxes(rank,shape)
       if HAS_TIMBA:
         _dprint(3, 'called setInitialSelectedAxes')
   # end StoreVellsData

   def isVellsScalar(self):
     """ returns true if no cells structure so data must be scalar """
     if HAS_TIMBA:
       _dprint(3,'returning self.scalar_data value ', self.scalar_data)
     return self.scalar_data

   def getScalarString(self):
     """ returns string of scalar values """
     return self.scalar_string

   def getShapeChange(self):
     """ returns true if data shape has changed """
     return self.shape_change

   def getNumPlanes(self):
     """ returns the number of vells planes that are stored """
     return self._number_of_planes

   def getMenuData(self):
     """ returns the labels for vells selection menu """
     if self.actual_rank > 1:
       return (self._menu_labels, self._perturbations_index, self._planes_index, self.dims,self._menu_labels_big)
     else:
       return (self._menu_labels, self._perturbations_index, self._planes_index, self.dims, None)


   def getVellsPlotLabels(self):
     """ returns the labels for vells plot menu """
     return self._plot_menu_labels

   def activePlaneHasFlags(self):
     """ returns True if active plane has associated flags array """
     key = "flag data " + str(self._active_plane)
     if key in self._plot_flags_dict:
       return True
     else:
       return False

   def getActiveFlagData(self):
     """ returns flag data associated with active plane """
     key = "flag data " + str(self._active_plane)
     if self.array_tuple is None:
       return self._plot_flags_dict[key]
     else:
       return self._plot_flags_dict[key][self.array_tuple]

   def getActivePlane(self):
     """ returns number of active plane """
     return self._active_plane

   def getPlotLabel(self):
     """ returns the plot label for a given plane or perturbed plane """
     key = ""
     if not self._active_perturb is None:
       tag = "] perturbed value "
       key =  "[" + str(self._active_plane) + tag + str(self._active_perturb) 
     else:
       key = "[" + str(self._active_plane) + "]" 
     return self._plot_labels[key]

   def getActivePerturbData(self):
     """ returns the vells data for the perturbed data associated with
         the active plane
     """
     tag = "] perturbed value "
     key =  "[" + str(self._active_plane) + tag + str(self._active_perturb) 
     return self._plot_vells_dict[key]

   def getActiveData(self):
     key = ""
     if not self._active_perturb is None:
       tag = "] perturbed value "
       key =  "[" + str(self._active_plane) + tag + str(self._active_perturb) 
     else:
       key = "[" + str(self._active_plane) + "]" 
     try:
       rank = self._plot_vells_dict[key].ndim
       shape = self._plot_vells_dict[key].shape
       if rank != self.rank or shape != self.shape:
         self.setInitialSelectedAxes (rank, shape)
     except:
       self.array_tuple = None
     if self.array_tuple is None:
       selected_array =  self._plot_vells_dict[key]
     else:
       if HAS_TIMBA:
         _dprint(3, 'self.array_tuple ',  self.array_tuple)
         _dprint(3, 'self._plot_vells_dict[key][self.array_tuple] has rank ',  self._plot_vells_dict[key][self.array_tuple].ndim)
         _dprint(3, 'self._plot_vells_dict[key][self.array_tuple] min and max: ', self._plot_vells_dict[key][self.array_tuple].min(), ' ', self._plot_vells_dict[key][self.array_tuple].max())
       selected_array =  self._plot_vells_dict[key][self.array_tuple]
     if self._request_full_image:
       self._request_full_image = False
       full_array = self.create_full_image(selected_array)
       return full_array
     else:
       return selected_array

   def request_full_image(self,signal):
     self._request_full_image = signal

   def getActivePlot(self):
     return self._active_plane
 
   def create_full_image(self, selected_image):
     # check if we may have to generate enlarged data set
     expected_size = []
     for k in range(len(self.axis_labels)):
       size = self.vells_axis_parms[self.axis_labels[k]][6]
       if self.array_tuple is None:
         if not size is None:
           expected_size.append(size)
       else:
         if self.array_tuple[k] == 0:
           expected_size.append(0)
         else:
           if not size is None:
             expected_size.append(size)
     full_array_tuple = tuple(expected_size)
     full_array = ones(full_array_tuple, selected_image.type())

     # create NaN
     try:
       full_array = full_array  * NAN
     except:
       pass

     orig_shape = selected_image.shape
     selection_list = []
     for k in range(len(orig_shape)):
       if orig_shape[k] != expected_size[k]:
         current_label = self.axis_labels[k] 
         grid_array = self.vells_axis_parms[current_label][4]
         delta = self.vells_axis_parms[current_label][5]
         selection_list.append(0)
       else:
         axis_slice = slice(0,orig_shape[k])
         selection_list.append(axis_slice)
     
     counter = 0
     expected_loc = grid_array[0]
     increment = delta[0]
     new_selection = []
     for k in range(len(selection_list)):
       new_selection.append(selection_list[k])
     for k in range(expected_size[1]):
       if expected_loc == grid_array[counter]:
         selection_list[1] = counter
         new_selection[1] = k
         tuple_selection_list = tuple(selection_list)
         tuple_new_selection = tuple(new_selection)
         try:
           full_array[tuple_new_selection] = selected_image[tuple_selection_list]
         except:
           pass
         counter = counter + 1
       expected_loc = expected_loc + increment
     return full_array
         
   def getActiveDataRanks(self):
     if HAS_TIMBA:
       _dprint(3, 'returning values self.actual_rank, self.rank, self.shape ', self.actual_rank, ' ', self.rank, ' ', self.shape)
     return (self.actual_rank, self.rank, self.shape)

   def setActivePlane(self, active_plane=0):
     self._active_plane = active_plane

   def set_3D_Display(self, display_3D=False):
     self.display_3D = display_3D
     self.initialSelection = False

   def setInitialSelection(self, selection_flag=False):
     self.initialSelection = selection_flag

   def setActivePerturb(self, active_perturb=0):
     self._active_perturb = active_perturb

   def getActiveAxisParms(self):
     return [self.first_axis_parm, self.second_axis_parm, self.third_axis_parm]

   def unravelMenuId(self, menuid=0):
      id_string = self._key_menu_labels[menuid] 
      self._active_perturb = None
      self._active_plane = 0
      perturb_loc = id_string.find("perturbed value")
      str_len = len(id_string)
      if perturb_loc >= 0:
        self._active_perturb = int(id_string[perturb_loc+15:str_len])
      
      request_plane_string = '0'
      plane_loc = id_string.find("[")
      if plane_loc >= 0:
        closing_bracket = id_string.find("]")
        if closing_bracket >= 0:
          request_plane_string = id_string[plane_loc+1:closing_bracket]
        self._active_plane = int(request_plane_string)

   def unsetSelectedAxes (self):
     self.array_tuple = None
     self.array_selector = []
 
   def updateArraySelector (self,lcd_number, slider_value):
    if len(self.array_selector) > 0:
      self.array_selector[lcd_number] = slider_value
      self.array_tuple = tuple(self.array_selector)

   def setInitialSelectedAxes (self, rank, shape, reset=False):
     try:
       if not reset and rank == self.rank and self.shape == shape:
         self.shape_change = False
         return
       else:
         self.array_selector = []
         self.array_tuple = None
         first_axis = None
         second_axis = None
         third_axis = None
         self.first_axis_parm = None
         self.second_axis_parm = None
         self.third_axis_parm = None
         self.actual_rank = 0
         if HAS_TIMBA:
           _dprint(3, 'self.actual_rank set ', self.actual_rank)
         self.rank = rank
         self.shape = shape
         self.shape_change = True
       if HAS_TIMBA:
         _dprint(3, 'rank ', rank)
         _dprint(3, 'shape ', shape)
         _dprint(3, 'self.axis_labels ', self.axis_labels)
         _dprint(3, 'self.shape_change ', self.shape_change)
       for i in range(rank-1,-1,-1):
         if HAS_TIMBA:
           _dprint(3, 'testing axes for shape[i] ', i, ' ', shape[i])
         if shape[i] > 1:
           self.actual_rank = self.actual_rank + 1
           if HAS_TIMBA:
             _dprint(3, 'self.actual rank now ', self.actual_rank)
         if shape[i] > 1 and self.display_3D and third_axis is None:
           third_axis = i
           self.third_axis_parm = self.axis_labels[i]
         elif shape[i] > 1 and second_axis is None:
           second_axis = i
           self.second_axis_parm = self.axis_labels[i]
           if HAS_TIMBA:
             _dprint(3, 'second axis becomes ', second_axis)
         elif shape[i] > 1 and first_axis is None:
           first_axis = i
           self.first_axis_parm = self.axis_labels[i]
           if HAS_TIMBA:
             _dprint(3, 'first axis becomes ', first_axis)
       if rank > 2:
         if not first_axis is None and not second_axis is None:
           self.array_selector = plot_func.create_array_selector(None, rank, shape, first_axis,second_axis,third_axis)
           self.array_tuple = tuple(self.array_selector)
         _dprint(3, 'array selector tuple ', self.array_tuple)
     except:
       _dprint(3, 'got an exception')
       self.array_selector = []
       self.array_tuple = None
     self.initialSelection = True
     if HAS_TIMBA:
       _dprint(3, 'self.first_axis_parm ', self.first_axis_parm)
       _dprint(3, 'self.second_axis_parm ', self.second_axis_parm)
       _dprint(3, 'self.third_axis_parm ', self.third_axis_parm)

   def setSelectedAxes (self,first_axis, second_axis, third_axis=-1):
     self.array_selector = []
     self.array_tuple = None
     try:
       rank = self.getActiveData().ndim
       if rank <= 2:
         return
       else:
         shape = self.getActiveData().shape
         self.array_selector = plot_func.create_array_selector(None, rank, shape, first_axis,second_axis,third_axis)
         self.array_tuple = tuple(self.array_selector)
         for i in range(rank):
           if i == first_axis:
             self.first_axis_parm = self.axis_labels[i]
           elif i == second_axis:
             self.second_axis_parm = self.axis_labels[i]
           elif i == third_axis:
             self.third_axis_parm = self.axis_labels[i]
     except:
       self.array_selector = []
       self.array_tuple = None

   def getDataRange(self):
     return [self.data_max, self.data_min]

   def setDataRange(self):
     """ figure out minima and maxima of active array """
     data_array = self.getActiveData()
     if self.test_scalar(data_array):
       self.data_min = data_array
       self.data_max = data_array
     else:
       if data_array.type() == Complex32 or data_array.type() == Complex64:
         real_array = data_array.real
         imag_array = data_array.imag
         real_min = real_array.min()
         real_max = real_array.max()
         imag_min = imag_array.min()
         imag_max = imag_array.max()
         if real_min < imag_min:
           self.data_min = real_min
         else:
           self.data_min = imag_min
         if real_max > imag_max:
           self.data_max = real_max
         else:
           self.data_max = imag_max
       else:
         self.data_min = data_array.min()
         self.data_max = data_array.max()

   def test_scalar(self, test_array):
     is_scalar = False
     try:
       shape = test_array.shape
     except:
       is_scalar = True
     return is_scalar

   
def main(args):
  print('we are in main') 

# Admire
if __name__ == '__main__':
    main(sys.argv)

